"""
EvEditor (Evennia Line Editor)

This implements an advanced line editor for editing longer texts in-game. The
editor mimics the command mechanisms of the "VI" editor (a famous line-by-line
editor) as far as reasonable.

Features of the editor:

- undo/redo.
- edit/replace on any line of the buffer.
- search&replace text anywhere in buffer.
- formatting of buffer, or selection, to certain width + indentations.
- allow to echo the input or not, depending on your client.
- in-built help

To use the editor, just import EvEditor from this module and initialize it:

```python
from evennia.utils.eveditor import EvEditor

# set up an editor to edit the caller's 'desc' Attribute
def _loadfunc(caller):
    return caller.db.desc

def _savefunc(caller, buffer):
    caller.db.desc = buffer.strip()
    return True

def _quitfunc(caller):
    caller.msg("Custom quit message")

# start the editor
EvEditor(caller, loadfunc=None, savefunc=None, quitfunc=None, key="",
         persistent=True, code=False)
```

The editor can also be used to format Python code and be made to
survive a reload. See the `EvEditor` class for more details.

"""
import re

from django.conf import settings
from django.utils.translation import gettext as _

from evennia import CmdSet
from evennia.commands import cmdhandler
from evennia.utils import dedent, fill, is_iter, justify, logger, to_str, utils
from evennia.utils.ansi import raw

# we use cmdhandler instead of evennia.syscmdkeys to
# avoid some cases of loading before evennia init'd
_CMD_NOMATCH = cmdhandler.CMD_NOMATCH
_CMD_NOINPUT = cmdhandler.CMD_NOINPUT

_RE_GROUP = re.compile(r"\".*?\"|\'.*?\'|\S*")
_COMMAND_DEFAULT_CLASS = utils.class_from_module(settings.COMMAND_DEFAULT_CLASS)
# use NAWS in the future?
_DEFAULT_WIDTH = settings.CLIENT_DEFAULT_WIDTH

# -------------------------------------------------------------
#
# texts
#
# -------------------------------------------------------------

_HELP_TEXT = _(
    """
 <txt>  - any non-command is appended to the end of the buffer.
 :  <l> - view buffer or only line(s) <l>
 :: <l> - raw-view buffer or only line(s) <l>
 :::    - escape - enter ':' as the only character on the line.
 :h     - this help.

 :w     - save the buffer (don't quit)
 :wq    - save buffer and quit
 :q     - quit (will be asked to save if buffer was changed)
 :q!    - quit without saving, no questions asked

 :u     - (undo) step backwards in undo history
 :uu    - (redo) step forward in undo history
 :UU    - reset all changes back to initial state

 :dd <l>     - delete last line or line(s) <l>
 :dw <l> <w> - delete word or regex <w> in entire buffer or on line <l>
 :DD         - clear entire buffer

 :y  <l>        - yank (copy) line(s) <l> to the copy buffer
 :x  <l>        - cut line(s) <l> and store it in the copy buffer
 :p  <l>        - put (paste) previously copied line(s) directly after <l>
 :i  <l> <txt>  - insert new text <txt> at line <l>. Old line will move down
 :r  <l> <txt>  - replace line <l> with text <txt>
 :I  <l> <txt>  - insert text at the beginning of line <l>
 :A  <l> <txt>  - append text after the end of line <l>

 :s <l> <w> <txt> - search/replace word or regex <w> in buffer or on line <l>

 :j <l> <w> - justify buffer or line <l>. <w> is f, c, l or r. Default f (full)
 :f <l>     - flood-fill entire buffer or line <l>: Equivalent to :j left
 :fi <l>    - indent entire buffer or line <l>
 :fd <l>    - de-indent entire buffer or line <l>

 :echo - turn echoing of the input on/off (helpful for some clients)
"""
)

_HELP_LEGEND = _(
    """
    Legend:
    <l>   - line number, like '5' or range, like '3:7'.
    <w>   - a single word, or multiple words with quotes around them.
    <txt> - longer string, usually not needing quotes.
"""
)

_HELP_CODE = _(
    """
 :!    - Execute code buffer without saving
 :<    - Decrease the level of automatic indentation for the next lines
 :>    - Increase the level of automatic indentation for the next lines
 :=    - Switch automatic indentation on/off
""".lstrip(
        "\n"
    )
)

_ERROR_LOADFUNC = _(
    """
{error}

|rBuffer load function error. Could not load initial data.|n
"""
)

_ERROR_SAVEFUNC = _(
    """
{error}

|rSave function returned an error. Buffer not saved.|n
"""
)

_ERROR_NO_SAVEFUNC = _("|rNo save function defined. Buffer cannot be saved.|n")

_MSG_SAVE_NO_CHANGE = _("No changes need saving")
_DEFAULT_NO_QUITFUNC = _("Exited editor.")

_ERROR_QUITFUNC = _(
    """
{error}

|rQuit function gave an error. Skipping.|n
"""
)

_ERROR_PERSISTENT_SAVING = _(
    """
{error}

|rThe editor state could not be saved for persistent mode. Switching
to non-persistent mode (which means the editor session won't survive
an eventual server reload - so save often!)|n
"""
)

_TRACE_PERSISTENT_SAVING = _(
    "EvEditor persistent-mode error. Commonly, this is because one or "
    "more of the EvEditor callbacks could not be pickled, for example "
    "because it's a class method or is defined inside another function."
)


_MSG_NO_UNDO = _("Nothing to undo.")
_MSG_NO_REDO = _("Nothing to redo.")
_MSG_UNDO = _("Undid one step.")
_MSG_REDO = _("Redid one step.")

# -------------------------------------------------------------
#
# Handle yes/no quit question
#
# -------------------------------------------------------------


class CmdSaveYesNo(_COMMAND_DEFAULT_CLASS):
    """
    Save the editor state on quit. This catches
    nomatches (defaults to Yes), and avoid saves only if
    command was given specifically as "no" or "n".
    """

    key = _CMD_NOMATCH
    aliases = _CMD_NOINPUT
    locks = "cmd:all()"
    help_cateogory = "LineEditor"

    def func(self):
        """
        Implement the yes/no choice.

        """
        # this is only called from inside the lineeditor
        # so caller.ndb._lineditor must be set.

        self.caller.cmdset.remove(SaveYesNoCmdSet)
        if self.raw_string.strip().lower() in ("no", "n"):
            # answered no
            self.caller.msg(self.caller.ndb._eveditor.quit())
        else:
            # answered yes (default)
            self.caller.ndb._eveditor.save_buffer()
            self.caller.ndb._eveditor.quit()


class SaveYesNoCmdSet(CmdSet):
    """
    Stores the yesno question

    """

    key = "quitsave_yesno"
    priority = 150  # override other cmdsets.
    mergetype = "Replace"

    def at_cmdset_creation(self):
        """at cmdset creation"""
        self.add(CmdSaveYesNo())


# -------------------------------------------------------------
#
# Editor commands
#
# -------------------------------------------------------------


class CmdEditorBase(_COMMAND_DEFAULT_CLASS):
    """
    Base parent for editor commands
    """

    locks = "cmd:all()"
    help_entry = "LineEditor"

    editor = None

    def parse(self):
        """
        Handles pre-parsing. Editor commands are on the form

        ::

            :cmd [li] [w] [txt]

        Where all arguments are optional.

        - `li`  - line number (int), starting from 1. This could also
              be a range given as <l>:<l>.
        - `w`  - word(s) (string), could be encased in quotes.
        - `txt` - extra text (string), could be encased in quotes.

        """

        editor = self.caller.ndb._eveditor
        if not editor:
            # this will completely replace the editor
            _load_editor(self.caller)
            editor = self.caller.ndb._eveditor
        self.editor = editor

        linebuffer = self.editor.get_buffer().split("\n")

        nlines = len(linebuffer)

        # The regular expression will split the line by whitespaces,
        # stripping extra whitespaces, except if the text is
        # surrounded by single- or double quotes, in which case they
        # will be kept together and extra whitespace preserved. You
        # can input quotes on the line by alternating single and
        # double quotes.
        arglist = [part for part in _RE_GROUP.findall(self.args) if part]
        temp = []
        for arg in arglist:
            # we want to clean the quotes, but only one type,
            # in case we are nesting.
            if arg.startswith('"'):
                arg.strip('"')
            elif arg.startswith("'"):
                arg.strip("'")
            temp.append(arg)
        arglist = temp

        # A dumb split, without grouping quotes
        words = self.args.split()

        # current line number
        cline = nlines - 1

        # the first argument could also be a range of line numbers, on the
        # form <lstart>:<lend>. Either of the ends could be missing, to
        # mean start/end of buffer respectively.

        lstart, lend = cline, cline + 1
        linerange = False
        if arglist and arglist[0].count(":") == 1:
            part1, part2 = arglist[0].split(":")
            if part1 and part1.isdigit():
                lstart = min(max(0, int(part1)) - 1, nlines)
                linerange = True
            if part2 and part2.isdigit():
                lend = min(lstart + 1, int(part2)) + 1
                linerange = True
        elif arglist and arglist[0].isdigit():
            lstart = min(max(0, int(arglist[0]) - 1), nlines)
            lend = lstart + 1
            linerange = True
        if linerange:
            arglist = arglist[1:]

        # nicer output formatting of the line range.
        lstr = (
            "line %i" % (lstart + 1)
            if not linerange or lstart + 1 == lend
            else "lines %i-%i" % (lstart + 1, lend)
        )

        # arg1 and arg2 is whatever arguments. Line numbers or -ranges are
        # never included here.
        args = " ".join(arglist)
        arg1, arg2 = "", ""
        if len(arglist) > 1:
            arg1, arg2 = arglist[0], " ".join(arglist[1:])
        else:
            arg1 = " ".join(arglist)

        # store for use in func()

        self.linebuffer = linebuffer
        self.nlines = nlines
        self.arglist = arglist
        self.cline = cline
        self.lstart = lstart
        self.lend = lend
        self.linerange = linerange
        self.lstr = lstr
        self.words = words
        self.args = args
        self.arg1 = arg1
        self.arg2 = arg2


def _load_editor(caller):
    """
    Load persistent editor from storage.

    """
    saved_options = caller.attributes.get("_eveditor_saved")
    saved_buffer, saved_undo = caller.attributes.get("_eveditor_buffer_temp", (None, None))
    unsaved = caller.attributes.get("_eveditor_unsaved", False)
    indent = caller.attributes.get("_eveditor_indent", 0)
    if saved_options:
        eveditor = EvEditor(caller, **saved_options[0])
        if saved_buffer:
            # we have to re-save the buffer data so we can handle subsequent restarts
            caller.attributes.add("_eveditor_buffer_temp", (saved_buffer, saved_undo))
            setattr(eveditor, "_buffer", saved_buffer)
            setattr(eveditor, "_undo_buffer", saved_undo)
            setattr(eveditor, "_undo_pos", len(saved_undo) - 1)
            setattr(eveditor, "_unsaved", unsaved)
            setattr(eveditor, "_indent", indent)
        for key, value in saved_options[1].items():
            setattr(eveditor, key, value)
    else:
        # something went wrong. Cleanup.
        caller.cmdset.remove(EvEditorCmdSet)


class CmdLineInput(CmdEditorBase):
    """
    No command match - Inputs line of text into buffer.

    """

    key = _CMD_NOMATCH
    aliases = _CMD_NOINPUT

    def func(self):
        """
        Adds the line without any formatting changes.

        If the editor handles code, it might add automatic
        indentation.
        """
        caller = self.caller
        editor = caller.ndb._eveditor
        buf = editor.get_buffer()

        # add a line of text to buffer
        line = self.raw_string.strip("\r\n")
        if editor._codefunc and editor._indent >= 0:
            # if automatic indentation is active, add spaces
            line = editor.deduce_indent(line, buf)
        buf = line if not buf else buf + "\n%s" % line
        self.editor.update_buffer(buf)
        if self.editor._echo_mode:
            # need to do it here or we will be off one line
            cline = len(self.editor.get_buffer().split("\n"))
            if editor._codefunc:
                # display the current level of identation
                indent = editor._indent
                if indent < 0:
                    indent = "off"

                self.caller.msg("|b%02i|||n (|g%s|n) %s" % (cline, indent, raw(line)))
            else:
                self.caller.msg("|b%02i|||n %s" % (cline, raw(self.args)))


class CmdEditorGroup(CmdEditorBase):
    """
    Commands for the editor
    """

    key = ":editor_command_group"
    aliases = [
        ":",
        "::",
        ":::",
        ":h",
        ":w",
        ":wq",
        ":q",
        ":q!",
        ":u",
        ":uu",
        ":UU",
        ":dd",
        ":dw",
        ":DD",
        ":y",
        ":x",
        ":p",
        ":i",
        ":j",
        ":r",
        ":I",
        ":A",
        ":s",
        ":S",
        ":f",
        ":fi",
        ":fd",
        ":echo",
        ":!",
        ":<",
        ":>",
        ":=",
    ]
    arg_regex = r"\s.*?|$"

    def func(self):
        """
        This command handles all the in-editor :-style commands. Since
        each command is small and very limited, this makes for a more
        efficient presentation.

        """
        caller = self.caller
        editor = caller.ndb._eveditor

        linebuffer = self.linebuffer
        lstart, lend = self.lstart, self.lend
        cmd = self.cmdstring
        echo_mode = self.editor._echo_mode

        if cmd == ":":
            # Echo buffer
            if self.linerange:
                buf = linebuffer[lstart:lend]
                editor.display_buffer(buf=buf, offset=lstart)
            else:
                editor.display_buffer()
        elif cmd == "::":
            # Echo buffer without the line numbers and syntax parsing
            if self.linerange:
                buf = linebuffer[lstart:lend]
                editor.display_buffer(buf=buf, offset=lstart, linenums=False, options={"raw": True})
            else:
                editor.display_buffer(linenums=False, options={"raw": True})
        elif cmd == ":::":
            # Insert single colon alone on a line
            editor.update_buffer([":"] if lstart == 0 else linebuffer + [":"])
            if echo_mode:
                caller.msg(_("Single ':' added to buffer."))
        elif cmd == ":h":
            # help entry
            editor.display_help()
        elif cmd == ":w":
            # save without quitting
            editor.save_buffer()
        elif cmd == ":wq":
            # save and quit
            editor.save_buffer()
            editor.quit()
        elif cmd == ":q":
            # quit. If not saved, will ask
            if self.editor._unsaved:
                caller.cmdset.add(SaveYesNoCmdSet)
                caller.msg(_("Save before quitting?") + " |lcyes|lt[Y]|le/|lcno|ltN|le")
            else:
                editor.quit()
        elif cmd == ":q!":
            # force quit, not checking saving
            editor.quit()
        elif cmd == ":u":
            # undo
            editor.update_undo(-1)
        elif cmd == ":uu":
            # redo
            editor.update_undo(1)
        elif cmd == ":UU":
            # reset buffer
            editor.update_buffer(editor._pristine_buffer)
            caller.msg(_("Reverted all changes to the buffer back to original state."))
        elif cmd == ":dd":
            # :dd <l> - delete line <l>
            buf = linebuffer[:lstart] + linebuffer[lend:]
            editor.update_buffer(buf)
            caller.msg(_("Deleted {string}.").format(string=self.lstr))
        elif cmd == ":dw":
            # :dw <w> - delete word in entire buffer
            # :dw <l> <w> delete word only on line(s) <l>
            if not self.arg1:
                caller.msg(_("You must give a search word to delete."))
            else:
                if not self.linerange:
                    lstart = 0
                    lend = self.cline + 1
                    caller.msg(
                        _("Removed {arg1} for lines {l1}-{l2}.").format(
                            arg1=self.arg1, l1=lstart + 1, l2=lend + 1
                        )
                    )
                else:
                    caller.msg(
                        _("Removed {arg1} for {line}.").format(arg1=self.arg1, line=self.lstr)
                    )
                sarea = "\n".join(linebuffer[lstart:lend])
                sarea = re.sub(r"%s" % self.arg1.strip("'").strip('"'), "", sarea, re.MULTILINE)
                buf = linebuffer[:lstart] + sarea.split("\n") + linebuffer[lend:]
                editor.update_buffer(buf)
        elif cmd == ":DD":
            # clear buffer
            editor.update_buffer("")

            # Reset indentation level to 0
            if editor._codefunc:
                if editor._indent >= 0:
                    editor._indent = 0
                    if editor._persistent:
                        caller.attributes.add("_eveditor_indent", 0)
            caller.msg(_("Cleared {nlines} lines from buffer.").format(nlines=self.nlines))
        elif cmd == ":y":
            # :y <l> - yank line(s) to copy buffer
            cbuf = linebuffer[lstart:lend]
            editor._copy_buffer = cbuf
            caller.msg(_("{line}, {cbuf} yanked.").format(line=self.lstr.capitalize(), cbuf=cbuf))
        elif cmd == ":x":
            # :x <l> - cut line to copy buffer
            cbuf = linebuffer[lstart:lend]
            editor._copy_buffer = cbuf
            buf = linebuffer[:lstart] + linebuffer[lend:]
            editor.update_buffer(buf)
            caller.msg(_("{line}, {cbuf} cut.").format(line=self.lstr.capitalize(), cbuf=cbuf))
        elif cmd == ":p":
            # :p <l> paste line(s) from copy buffer
            if not editor._copy_buffer:
                caller.msg(_("Copy buffer is empty."))
            else:
                buf = linebuffer[:lstart] + editor._copy_buffer + linebuffer[lstart:]
                editor.update_buffer(buf)
                caller.msg(
                    _("Pasted buffer {cbuf} to {line}.").format(
                        cbuf=editor._copy_buffer, line=self.lstr
                    )
                )
        elif cmd == ":i":
            # :i <l> <txt> - insert new line
            new_lines = self.args.split("\n")
            if not new_lines:
                caller.msg(_("You need to enter a new line and where to insert it."))
            else:
                buf = linebuffer[:lstart] + new_lines + linebuffer[lstart:]
                editor.update_buffer(buf)
                caller.msg(
                    _("Inserted {num} new line(s) at {line}.").format(
                        num=len(new_lines), line=self.lstr
                    )
                )
        elif cmd == ":r":
            # :r <l> <txt> - replace lines
            new_lines = self.args.split("\n")
            if not new_lines:
                caller.msg(_("You need to enter a replacement string."))
            else:
                buf = linebuffer[:lstart] + new_lines + linebuffer[lend:]
                editor.update_buffer(buf)
                caller.msg(
                    _("Replaced {num} line(s) at {line}.").format(
                        num=len(new_lines), line=self.lstr
                    )
                )
        elif cmd == ":I":
            # :I <l> <txt> - insert text at beginning of line(s) <l>
            if not self.raw_string and not editor._codefunc:
                caller.msg(_("You need to enter text to insert."))
            else:
                buf = (
                    linebuffer[:lstart]
                    + ["%s%s" % (self.args, line) for line in linebuffer[lstart:lend]]
                    + linebuffer[lend:]
                )
                editor.update_buffer(buf)
                caller.msg(_("Inserted text at beginning of {line}.").format(line=self.lstr))
        elif cmd == ":A":
            # :A <l> <txt> - append text after end of line(s)
            if not self.args:
                caller.msg(_("You need to enter text to append."))
            else:
                buf = (
                    linebuffer[:lstart]
                    + ["%s%s" % (line, self.args) for line in linebuffer[lstart:lend]]
                    + linebuffer[lend:]
                )
                editor.update_buffer(buf)
                caller.msg(_("Appended text to end of {line}.").format(line=self.lstr))
        elif cmd == ":s":
            # :s <li> <w> <txt> - search and replace words
            # in entire buffer or on certain lines
            if not self.arg1 or not self.arg2:
                caller.msg(_("You must give a search word and something to replace it with."))
            else:
                if not self.linerange:
                    lstart = 0
                    lend = self.cline + 1
                    caller.msg(
                        _("Search-replaced {arg1} -> {arg2} for lines {l1}-{l2}.").format(
                            arg1=self.arg1, arg2=self.arg2, l1=lstart + 1, l2=lend
                        )
                    )
                else:
                    caller.msg(
                        _("Search-replaced {arg1} -> {arg2} for {line}.").format(
                            arg1=self.arg1, arg2=self.arg2, line=self.lstr
                        )
                    )
                sarea = "\n".join(linebuffer[lstart:lend])

                regex = r"%s|^%s(?=\s)|(?<=\s)%s(?=\s)|^%s$|(?<=\s)%s$"
                regarg = self.arg1.strip("'").strip('"')
                if " " in regarg:
                    regarg = regarg.replace(" ", " +")
                sarea = re.sub(
                    regex % (regarg, regarg, regarg, regarg, regarg),
                    self.arg2.strip("'").strip('"'),
                    sarea,
                    re.MULTILINE,
                )
                buf = linebuffer[:lstart] + sarea.split("\n") + linebuffer[lend:]
                editor.update_buffer(buf)
        elif cmd == ":f":
            # :f <l> flood-fill buffer or <l> lines of buffer.
            width = _DEFAULT_WIDTH
            if not self.linerange:
                lstart = 0
                lend = self.cline + 1
                caller.msg(_("Flood filled lines {l1}-{l2}.").format(l1=lstart + 1, l2=lend))
            else:
                caller.msg(_("Flood filled {line}.").format(line=self.lstr))
            fbuf = "\n".join(linebuffer[lstart:lend])
            fbuf = fill(fbuf, width=width)
            buf = linebuffer[:lstart] + fbuf.split("\n") + linebuffer[lend:]
            editor.update_buffer(buf)
        elif cmd == ":j":
            # :f <l> <w>  justify buffer of <l> with <w> as align (one of
            # f(ull), c(enter), r(ight) or l(left). Default is full.
            align_map = {
                "full": "f",
                "f": "f",
                "center": "c",
                "c": "c",
                "right": "r",
                "r": "r",
                "left": "l",
                "l": "l",
            }
            align_name = {"f": "Full", "c": "Center", "l": "Left", "r": "Right"}
            width = _DEFAULT_WIDTH
            if self.arg1 and self.arg1.lower() not in align_map:
                self.caller.msg(
                    _("Valid justifications are")
                    + " [f]ull (default), [c]enter, [r]right or [l]eft"
                )
                return
            align = align_map[self.arg1.lower()] if self.arg1 else "f"
            if not self.linerange:
                lstart = 0
                lend = self.cline + 1
                self.caller.msg(
                    _("{align}-justified lines {l1}-{l2}.").format(
                        align=align_name[align], l1=lstart + 1, l2=lend
                    )
                )
            else:
                self.caller.msg(
                    _("{align}-justified {line}.").format(align=align_name[align], line=self.lstr)
                )
            jbuf = "\n".join(linebuffer[lstart:lend])
            jbuf = justify(jbuf, width=width, align=align)
            buf = linebuffer[:lstart] + jbuf.split("\n") + linebuffer[lend:]
            editor.update_buffer(buf)
        elif cmd == ":fi":
            # :fi <l> indent buffer or lines <l> of buffer.
            indent = " " * 4
            if not self.linerange:
                lstart = 0
                lend = self.cline + 1
                caller.msg(_("Indented lines {l1}-{l2}.").format(l1=lstart + 1, l2=lend))
            else:
                caller.msg(_("Indented {line}.").format(line=self.lstr))
            fbuf = [indent + line for line in linebuffer[lstart:lend]]
            buf = linebuffer[:lstart] + fbuf + linebuffer[lend:]
            editor.update_buffer(buf)
        elif cmd == ":fd":
            # :fi <l> indent buffer or lines <l> of buffer.
            if not self.linerange:
                lstart = 0
                lend = self.cline + 1
                caller.msg(
                    _("Removed left margin (dedented) lines {l1}-{l2}.").format(
                        l1=lstart + 1, l2=lend
                    )
                )
            else:
                caller.msg(_("Removed left margin (dedented) {line}.").format(line=self.lstr))
            fbuf = "\n".join(linebuffer[lstart:lend])
            fbuf = dedent(fbuf)
            buf = linebuffer[:lstart] + fbuf.split("\n") + linebuffer[lend:]
            editor.update_buffer(buf)
        elif cmd == ":echo":
            # set echoing on/off
            editor._echo_mode = not editor._echo_mode
            caller.msg(_("Echo mode set to {mode}").format(mode=editor._echo_mode))
        elif cmd == ":!":
            if editor._codefunc:
                editor._codefunc(caller, editor._buffer)
            else:
                caller.msg(_("This command is only available in code editor mode."))
        elif cmd == ":<":
            # :<
            if editor._codefunc:
                editor.decrease_indent()
                indent = editor._indent
                if indent >= 0:
                    caller.msg(
                        _("Decreased indentation: new indentation is {indent}.").format(
                            indent=indent
                        )
                    )
                else:
                    caller.msg(_("|rManual indentation is OFF.|n Use := to turn it on."))
            else:
                caller.msg(_("This command is only available in code editor mode."))
        elif cmd == ":>":
            # :>
            if editor._codefunc:
                editor.increase_indent()
                indent = editor._indent
                if indent >= 0:
                    caller.msg(
                        _("Increased indentation: new indentation is {indent}.").format(
                            indent=indent
                        )
                    )
                else:
                    caller.msg(_("|rManual indentation is OFF.|n Use := to turn it on."))
            else:
                caller.msg(_("This command is only available in code editor mode."))
        elif cmd == ":=":
            # :=
            if editor._codefunc:
                editor.swap_autoindent()
                indent = editor._indent
                if indent >= 0:
                    caller.msg(_("Auto-indentation turned on."))
                else:
                    caller.msg(_("Auto-indentation turned off."))
            else:
                caller.msg(_("This command is only available in code editor mode."))


class EvEditorCmdSet(CmdSet):
    """CmdSet for the editor commands"""

    key = "editorcmdset"
    mergetype = "Replace"

    def at_cmdset_creation(self):
        self.add(CmdLineInput())
        self.add(CmdEditorGroup())


# -------------------------------------------------------------
#
# Main Editor object
#
# -------------------------------------------------------------


class EvEditor:
    """
    This defines a line editor object. It creates all relevant commands
    and tracks the current state of the buffer. It also cleans up after
    itself.

    """

    def __init__(
        self,
        caller,
        loadfunc=None,
        savefunc=None,
        quitfunc=None,
        key="",
        persistent=False,
        codefunc=False,
    ):
        """
        Launches a full in-game line editor, mimicking the functionality of VIM.

        Args:
            caller (Object): Who is using the editor.
            loadfunc (callable, optional): This will be called as
                `loadfunc(caller)` when the editor is first started. Its
                return will be used as the editor's starting buffer.
            savefunc (callable, optional): This will be called as
                `savefunc(caller, buffer)` when the save-command is given and
                is used to actually determine where/how result is saved.
                It should return `True` if save was successful and also
                handle any feedback to the user.
            quitfunc (callable, optional): This will optionally be
                called as `quitfunc(caller)` when the editor is
                exited. If defined, it should handle all wanted feedback
                to the user.
            quitfunc_args (tuple, optional): Optional tuple of arguments to
                supply to `quitfunc`.
            key (str, optional): An optional key for naming this
                session and make it unique from other editing sessions.
            persistent (bool, optional): Make the editor survive a reboot. Note
                that if this is set, all callables must be possible to pickle
            codefunc (bool, optional): If given, will run the editor in code mode.
                This will be called as `codefunc(caller, buf)`.

        Notes:
            In persistent mode, all the input callables (savefunc etc)
            must be possible to be *pickled*, this excludes e.g.
            callables that are class methods or functions defined
            dynamically or as part of another function. In
            non-persistent mode no such restrictions exist.



        """
        self._key = key
        self._caller = caller
        self._caller.ndb._eveditor = self
        self._buffer = ""
        self._unsaved = False
        self._persistent = persistent
        self._indent = 0

        if loadfunc:
            self._loadfunc = loadfunc
        else:
            self._loadfunc = lambda caller: self._buffer
        self.load_buffer()
        if savefunc:
            self._savefunc = savefunc
        else:
            self._savefunc = lambda caller, buffer: caller.msg(_ERROR_NO_SAVEFUNC)
        if quitfunc:
            self._quitfunc = quitfunc
        else:
            self._quitfunc = lambda caller: caller.msg(_DEFAULT_NO_QUITFUNC)
        self._codefunc = codefunc

        # store the original version
        self._pristine_buffer = self._buffer
        self._sep = "-"

        # undo operation buffer
        self._undo_buffer = [self._buffer]
        self._undo_pos = 0
        self._undo_max = 20

        # copy buffer
        self._copy_buffer = []

        if persistent:
            # save in tuple {kwargs, other options}
            try:
                caller.attributes.add(
                    "_eveditor_saved",
                    (
                        dict(
                            loadfunc=loadfunc,
                            savefunc=savefunc,
                            quitfunc=quitfunc,
                            codefunc=codefunc,
                            key=key,
                            persistent=persistent,
                        ),
                        dict(_pristine_buffer=self._pristine_buffer, _sep=self._sep),
                    ),
                )
                caller.attributes.add("_eveditor_buffer_temp", (self._buffer, self._undo_buffer))
                caller.attributes.add("_eveditor_unsaved", False)
                caller.attributes.add("_eveditor_indent", 0)
            except Exception as err:
                caller.msg(_ERROR_PERSISTENT_SAVING.format(error=err))
                logger.log_trace(_TRACE_PERSISTENT_SAVING)
                persistent = False

        # Create the commands we need
        caller.cmdset.add(EvEditorCmdSet, persistent=persistent)

        # echo inserted text back to caller
        self._echo_mode = True

        # show the buffer ui
        self.display_buffer()

    def load_buffer(self):
        """
        Load the buffer using the load function hook.

        """
        try:
            self._buffer = self._loadfunc(self._caller)
            if not isinstance(self._buffer, str):
                self._caller.msg(
                    f"|rBuffer is of type |w{type(self._buffer)})|r. "
                    "Continuing, it is converted to a string "
                    "(and will be saved as such)!|n"
                )
                self._buffer = to_str(self._buffer)
        except Exception as e:
            from evennia.utils import logger

            logger.log_trace()
            self._caller.msg(_ERROR_LOADFUNC.format(error=e))

    def get_buffer(self):
        """
        Return:
            buffer (str): The current buffer.

        """
        return self._buffer

    def update_buffer(self, buf):
        """
        This should be called when the buffer has been changed
        somehow.  It will handle unsaved flag and undo updating.

        Args:
            buf (str): The text to update the buffer with.

        """
        if is_iter(buf):
            buf = "\n".join(buf)

        if buf != self._buffer:
            self._buffer = buf
            self.update_undo()
            self._unsaved = True
            if self._persistent:
                self._caller.attributes.add(
                    "_eveditor_buffer_temp", (self._buffer, self._undo_buffer)
                )
                self._caller.attributes.add("_eveditor_unsaved", True)
                self._caller.attributes.add("_eveditor_indent", self._indent)

    def quit(self):
        """
        Cleanly exit the editor.

        """
        try:
            self._quitfunc(self._caller)
        except Exception as e:
            self._caller.msg(_ERROR_QUITFUNC.format(error=e))
        self._caller.nattributes.remove("_eveditor")
        self._caller.attributes.remove("_eveditor_buffer_temp")
        self._caller.attributes.remove("_eveditor_saved")
        self._caller.attributes.remove("_eveditor_unsaved")
        self._caller.attributes.remove("_eveditor_indent")
        self._caller.cmdset.remove(EvEditorCmdSet)

    def save_buffer(self):
        """
        Saves the content of the buffer.

        """
        if self._unsaved or self._codefunc:
            # always save code - this allows us to tie execution to
            # saving if we want.
            try:
                if self._savefunc(self._caller, self._buffer):
                    # Save codes should return a true value to indicate
                    # save worked. The saving function is responsible for
                    # any status messages.
                    self._unsaved = False
            except Exception as e:
                self._caller.msg(_ERROR_SAVEFUNC.format(error=e))
        else:
            self._caller.msg(_MSG_SAVE_NO_CHANGE)

    def update_undo(self, step=None):
        """
        This updates the undo position.

        Args:
            step (int, optional): The amount of steps
                to progress the undo position to. This
                may be a negative value for undo and
                a positive value for redo.

        """
        if step and step < 0:
            # undo
            if self._undo_pos <= 0:
                self._caller.msg(_MSG_NO_UNDO)
            else:
                self._undo_pos = max(0, self._undo_pos + step)
                self._buffer = self._undo_buffer[self._undo_pos]
                self._caller.msg(_MSG_UNDO)
        elif step and step > 0:
            # redo
            if self._undo_pos >= len(self._undo_buffer) - 1 or self._undo_pos + 1 >= self._undo_max:
                self._caller.msg(_MSG_NO_REDO)
            else:
                self._undo_pos = min(
                    self._undo_pos + step, min(len(self._undo_buffer), self._undo_max) - 1
                )
                self._buffer = self._undo_buffer[self._undo_pos]
                self._caller.msg(_MSG_REDO)
        if not self._undo_buffer or (
            self._undo_buffer and self._buffer != self._undo_buffer[self._undo_pos]
        ):
            # save undo state
            self._undo_buffer = self._undo_buffer[: self._undo_pos + 1] + [self._buffer]
            self._undo_pos = len(self._undo_buffer) - 1

    def display_buffer(self, buf=None, offset=0, linenums=True, options={"raw": False}):
        """
        This displays the line editor buffer, or selected parts of it.

        Args:
            buf (str, optional): The buffer or part of buffer to display.
            offset (int, optional): If `buf` is set and is not the full buffer,
                `offset` should define the actual starting line number, to
                get the linenum display right.
            linenums (bool, optional): Show line numbers in buffer.
            options: raw (bool, optional): Tell protocol to not parse
                formatting information.

        """
        if buf is None:
            buf = self._buffer
        if is_iter(buf):
            buf = "\n".join(buf)

        lines = buf.split("\n")
        nlines = len(lines)
        nwords = len(buf.split())
        nchars = len(buf)

        sep = self._sep
        header = (
            "|n"
            + sep * 10
            + _("Line Editor [{name}]").format(name=self._key)
            + sep * (_DEFAULT_WIDTH - 24 - len(self._key))
        )
        footer = (
            "|n"
            + sep * 10
            + "[l:%02i w:%03i c:%04i]" % (nlines, nwords, nchars)
            + sep * 12
            + _("(:h for help)")
            + sep * (_DEFAULT_WIDTH - 54)
        )
        if linenums:
            main = "\n".join(
                "|b%02i|||n %s" % (iline + 1 + offset, raw(line))
                for iline, line in enumerate(lines)
            )
        else:
            main = "\n".join([raw(line) for line in lines])
        string = "%s\n%s\n%s" % (header, main, footer)
        self._caller.msg(string, options=options)

    def display_help(self):
        """
        Shows the help entry for the editor.

        """
        string = self._sep * _DEFAULT_WIDTH + _HELP_TEXT
        if self._codefunc:
            string += _HELP_CODE
        string += _HELP_LEGEND + self._sep * _DEFAULT_WIDTH
        self._caller.msg(string)

    def deduce_indent(self, line, buffer):
        """
        Try to deduce the level of indentation of the given line.

        """
        keywords = {
            "elif ": ["if "],
            "else:": ["if ", "try"],
            "except": ["try:"],
            "finally:": ["try:"],
        }
        opening_tags = ("if ", "try:", "for ", "while ")

        # If the line begins by one of the given keywords
        indent = self._indent
        if any(line.startswith(kw) for kw in keywords.keys()):
            # Get the keyword and matching begin tags
            keyword = [kw for kw in keywords if line.startswith(kw)][0]
            begin_tags = keywords[keyword]
            for oline in reversed(buffer.splitlines()):
                if any(oline.lstrip(" ").startswith(tag) for tag in begin_tags):
                    # This line begins with a begin tag, takes the identation
                    indent = (len(oline) - len(oline.lstrip(" "))) / 4
                    break

            self._indent = indent + 1
            if self._persistent:
                self._caller.attributes.add("_eveditor_indent", self._indent)
        elif any(line.startswith(kw) for kw in opening_tags):
            self._indent = indent + 1
            if self._persistent:
                self._caller.attributes.add("_eveditor_indent", self._indent)

        line = " " * 4 * indent + line
        return line

    def decrease_indent(self):
        """Decrease automatic indentation by 1 level."""
        if self._codefunc and self._indent > 0:
            self._indent -= 1
            if self._persistent:
                self._caller.attributes.add("_eveditor_indent", self._indent)

    def increase_indent(self):
        """Increase automatic indentation by 1 level."""
        if self._codefunc and self._indent >= 0:
            self._indent += 1
            if self._persistent:
                self._caller.attributes.add("_eveditor_indent", self._indent)

    def swap_autoindent(self):
        """Swap automatic indentation on or off."""
        if self._codefunc:
            if self._indent >= 0:
                self._indent = -1
            else:
                self._indent = 0

            if self._persistent:
                self._caller.attributes.add("_eveditor_indent", self._indent)



"""
EvAdventure Base combat utilities.

This establishes the basic building blocks for combat:

- `CombatFailure` - exception for combat-specific errors.
- `CombatAction` (and subclasses) - classes encompassing all the working around an action.
  They are initialized from 'action-dicts` - dictionaries with all the relevant data for the
  particular invocation
- `CombatHandler` - base class for running a combat. Exactly how this is used depends on the
  type of combat intended (twitch- or turn-based) so many details of this will be implemented
  in child classes.

----

"""

from evennia.scripts.scripts import DefaultScript
from evennia.typeclasses.attributes import AttributeProperty
from evennia.utils import evtable
from evennia.utils.create import create_script

from . import rules


class CombatFailure(RuntimeError):
    """
    Some failure during combat actions.

    """


class CombatAction:
    """
    Parent class for all actions.

    This represents the executable code to run to perform an action. It is initialized from an
    'action-dict', a set of properties stored in the action queue by each combatant.

    """

    def __init__(self, combathandler, combatant, action_dict):
        """
        Each key-value pair in the action-dict is stored as a property on this class
        for later access.

        Args:
            combatant (EvAdventureCharacter, EvAdventureNPC): The combatant performing
                the action.
            action_dict (dict): A dict containing all properties to initialize on this
                class. This should not be any keys with `_` prefix, since these are
                used internally by the class.

        """
        self.combathandler = combathandler
        self.combatant = combatant

        # store the action dicts' keys as properties accessible as e.g. action.target etc
        for key, val in action_dict.items():
            if not key.startswith("_"):
                setattr(self, key, val)

    def msg(self, message, broadcast=True):
        """
        Convenience route to the combathandler msg-sender mechanism.

        Args:
            message (str): Message to send; use `$You()` and `$You(other.key)` to refer to
                the combatant doing the action and other combatants, respectively.

        """
        self.combathandler.msg(message, combatant=self.combatant, broadcast=broadcast)

    def can_use(self):
        """
        Called to determine if the action is usable with the current settings. This does not
        actually perform the action.

        Returns:
            bool: If this action can be used at this time.

        """
        return True

    def execute(self):
        """
        Perform the action as the combatant. Should normally make use of the properties
        stored on the class during initialization.

        """
        pass

    def post_execute(self):
        """
        Called after execution.
        """
        pass


class CombatActionHold(CombatAction):
    """
    Action that does nothing.
    ::
        action_dict = {
                "key": "hold"
            }
    """


class CombatActionAttack(CombatAction):
    """
    A regular attack, using a wielded weapon.
    ::
        action-dict = {
                "key": "attack",
                "target": Character/Object
            }
    """

    def execute(self):
        attacker = self.combatant
        weapon = attacker.weapon
        target = self.target

        if weapon.at_pre_use(attacker, target):
            weapon.use(
                attacker, target, advantage=self.combathandler.has_advantage(attacker, target)
            )
            weapon.at_post_use(attacker, target)


class CombatActionStunt(CombatAction):
    """
    Perform a stunt the grants a beneficiary (can be self) advantage on their next action against a
    target. Whenever performing a stunt that would affect another negatively (giving them
    disadvantage against an ally, or granting an advantage against them, we need to make a check
    first. We don't do a check if giving an advantage to an ally or ourselves.
    ::
        action_dict = {
               "key": "stunt",
               "recipient": Character/NPC,
               "target": Character/NPC,
               "advantage": bool,  # if False, it's a disadvantage
               "stunt_type": Ability,  # what ability (like STR, DEX etc) to use to perform this stunt.
               "defense_type": Ability, # what ability to use to defend against (negative) effects of
                this stunt.
            }

    """

    def execute(self):
        combathandler = self.combathandler
        attacker = self.combatant
        recipient = self.recipient  # the one to receive the effect of the stunt
        target = self.target  # the affected by the stunt (can be the same as recipient/combatant)
        txt = ""

        if recipient == target:
            # grant another entity dis/advantage against themselves
            defender = recipient
        else:
            # recipient not same as target; who will defend depends on disadvantage or advantage
            # to give.
            defender = target if self.advantage else recipient

        # trying to give advantage to recipient against target. Target defends against caller
        is_success, _, txt = rules.dice.opposed_saving_throw(
            attacker,
            defender,
            attack_type=self.stunt_type,
            defense_type=self.defense_type,
            advantage=combathandler.has_advantage(attacker, defender),
            disadvantage=combathandler.has_disadvantage(attacker, defender),
        )

        self.msg(f"$You() $conj(attempt) stunt on $You({defender.key}). {txt}")

        # deal with results
        if is_success:
            if self.advantage:
                combathandler.give_advantage(recipient, target)
            else:
                combathandler.give_disadvantage(recipient, target)
            if recipient == self.combatant:
                self.msg(
                    f"$You() $conj(gain) {'advantage' if self.advantage else 'disadvantage'} "
                    f"against $You({target.key})!"
                )
            else:
                self.msg(
                    f"$You() $conj(cause) $You({recipient.key}) "
                    f"to gain {'advantage' if self.advantage else 'disadvantage'} "
                    f"against $You({target.key})!"
                )
        else:
            self.msg(f"$You({defender.key}) $conj(resist)! $You() $conj(fail) the stunt.")


class CombatActionUseItem(CombatAction):
    """
    Use an item in combat. This is meant for one-off or limited-use items (so things like
    scrolls and potions, not swords and shields). If this is some sort of weapon or spell rune,
    we refer to the item to determine what to use for attack/defense rolls.
    ::
        action_dict = {
                "key": "use",
                "item": Object
                "target": Character/NPC/Object/None
            }
    """

    def execute(self):
        item = self.item
        user = self.combatant
        target = self.target

        if item.at_pre_use(user, target):
            item.use(
                user,
                target,
                advantage=self.combathandler.has_advantage(user, target),
                disadvantage=self.combathandler.has_disadvantage(user, target),
            )
            item.at_post_use(user, target)


class CombatActionWield(CombatAction):
    """
    Wield a new weapon (or spell) from your inventory. This will swap out the one you are currently
    wielding, if any.
    ::
        action_dict = {
                "key": "wield",
                "item": Object
            }
    """

    def execute(self):
        self.combatant.equipment.move(self.item)
        self.msg(f"$You() $conj(wield) $You({self.item.key}).")


# main combathandler


class EvAdventureCombatBaseHandler(DefaultScript):
    """
    This script is created when a combat starts. It 'ticks' the combat and tracks
    all sides of it.

    """

    # available actions in combat
    action_classes = {
        "hold": CombatActionHold,
        "attack": CombatActionAttack,
        "stunt": CombatActionStunt,
        "use": CombatActionUseItem,
        "wield": CombatActionWield,
    }

    # fallback action if not selecting anything
    fallback_action_dict = AttributeProperty({"key": "hold"}, autocreate=False)

    @classmethod
    def get_or_create_combathandler(cls, obj, **kwargs):
        """
        Get or create a combathandler on `obj`.

        Args:
            obj (any): The Typeclassed entity to store the CombatHandler Script on. This could be
                a location (for turn-based combat) or a Character (for twitch-based combat).
        Keyword Args:
            combathandler_key (str): They key name for the script. Will be 'combathandler' by
                default.
            **kwargs: Arguments to the Script, if it is created.

        """
        if not obj:
            raise CombatFailure("Cannot start combat without a place to do it!")

        combathandler_key = kwargs.pop("key", "combathandler")
        combathandler = obj.ndb.combathandler
        if not combathandler or not combathandler.id:
            combathandler = obj.scripts.get(combathandler_key).first()
            if not combathandler:
                # have to create from scratch
                persistent = kwargs.pop("persistent", True)
                combathandler = create_script(
                    cls,
                    key=combathandler_key,
                    obj=obj,
                    persistent=persistent,
                    autostart=False,
                    **kwargs,
                )
            obj.ndb.combathandler = combathandler
        return combathandler

    def msg(self, message, combatant=None, broadcast=True, location=None):
        """
        Central place for sending messages to combatants. This allows
        for adding any combat-specific text-decoration in one place.

        Args:
            message (str): The message to send.
            combatant (Object): The 'You' in the message, if any.
            broadcast (bool): If `False`, `combatant` must be included and
                will be the only one to see the message. If `True`, send to
                everyone in the location.
            location (Object, optional): If given, use this as the location to
                send broadcast messages to. If not, use `self.obj` as that
                location.

        Notes:
            If `combatant` is given, use `$You/you()` markup to create
            a message that looks different depending on who sees it. Use
            `$You(combatant_key)` to refer to other combatants.

        """
        if not location:
            location = self.obj

        location_objs = location.contents

        exclude = []
        if not broadcast and combatant:
            exclude = [obj for obj in location_objs if obj is not combatant]

        location.msg_contents(
            message,
            exclude=exclude,
            from_obj=combatant,
            mapping={locobj.key: locobj for locobj in location_objs},
        )

    def get_combat_summary(self, combatant):
        """
        Get a 'battle report' - an overview of the current state of combat from the perspective
        of one of the sides.

        Args:
            combatant (EvAdventureCharacter, EvAdventureNPC): The combatant to get.

        Returns:
            EvTable: A table representing the current state of combat.

        Example:
        ::

                                        Goblin shaman (Perfect)
        Gregor (Hurt)                   Goblin brawler(Hurt)
        Bob (Perfect)         vs        Goblin grunt 1 (Hurt)
                                        Goblin grunt 2 (Perfect)
                                        Goblin grunt 3 (Wounded)

        """
        allies, enemies = self.get_sides(combatant)
        nallies, nenemies = len(allies), len(enemies)

        # prepare colors and hurt-levels
        allies = [f"{ally} ({ally.hurt_level})" for ally in allies]
        enemies = [f"{enemy} ({enemy.hurt_level})" for enemy in enemies]

        # the center column with the 'vs'
        vs_column = ["" for _ in range(max(nallies, nenemies))]
        vs_column[len(vs_column) // 2] = "|wvs|n"

        # the two allies / enemies columns should be centered vertically
        diff = abs(nallies - nenemies)
        top_empty = diff // 2
        bot_empty = diff - top_empty
        topfill = ["" for _ in range(top_empty)]
        botfill = ["" for _ in range(bot_empty)]

        if nallies >= nenemies:
            enemies = topfill + enemies + botfill
        else:
            allies = topfill + allies + botfill

        # make a table with three columns
        return evtable.EvTable(
            table=[
                evtable.EvColumn(*allies, align="l"),
                evtable.EvColumn(*vs_column, align="c"),
                evtable.EvColumn(*enemies, align="r"),
            ],
            border=None,
            maxwidth=78,
        )

    def get_sides(self, combatant):
        """
        Get a listing of the two 'sides' of this combat, from the perspective of the provided
        combatant. The sides don't need to be balanced.

        Args:
            combatant (Character or NPC): The one whose sides are to determined.

        Returns:
            tuple: A tuple of lists `(allies, enemies)`, from the perspective of `combatant`.

        Note:
            The sides are found by checking PCs vs NPCs. PCs can normally not attack other PCs, so
            are naturally allies. If the current room has the `allow_pvp` Attribute set, then _all_
            other combatants (PCs and NPCs alike) are considered valid enemies (one could expand
            this with group mechanics).

        """
        raise NotImplementedError

    def give_advantage(self, recipient, target):
        """
        Let a benefiter gain advantage against the target.

        Args:
            recipient (Character or NPC): The one to gain the advantage. This may or may not
                be the same entity that creates the advantage in the first place.
            target (Character or NPC): The one against which the target gains advantage. This
                could (in principle) be the same as the benefiter (e.g. gaining advantage on
                some future boost)

        """
        raise NotImplementedError

    def give_disadvantage(self, recipient, target):
        """
        Let an affected party gain disadvantage against a target.

        Args:
            recipient (Character or NPC): The one to get the disadvantage.
            target (Character or NPC): The one against which the target gains disadvantage, usually
            an enemy.

        """
        raise NotImplementedError

    def has_advantage(self, combatant, target):
        """
        Check if a given combatant has advantage against a target.

        Args:
            combatant (Character or NPC): The one to check if they have advantage
            target (Character or NPC): The target to check advantage against.

        """
        raise NotImplementedError

    def has_disadvantage(self, combatant, target):
        """
        Check if a given combatant has disadvantage against a target.

        Args:
            combatant (Character or NPC): The one to check if they have disadvantage
            target (Character or NPC): The target to check disadvantage against.

        """
        raise NotImplementedError

    def queue_action(self, action_dict, combatant=None):
        """
        Queue an action by adding the new actiondict.

        Args:
            action_dict (dict): A dict describing the action class by name along with properties.
            combatant (EvAdventureCharacter, EvAdventureNPC, optional): A combatant queueing the
                action.

        """
        raise NotImplementedError

    def execute_next_action(self, combatant):
        """
        Perform a combatant's next action.

        Args:
            combatant (EvAdventureCharacter, EvAdventureNPC): The combatant performing and action.


        """
        raise NotImplementedError

    def start_combat(self):
        """
        Start combat.

        """
        raise NotImplementedError

    def check_stop_combat(self):
        """
        Check if this combat should be aborted, whatever this means for the particular
        the particular combat type.

        Keyword Args:
            kwargs: Any extra keyword args used.

        Returns:
            bool: If `True`, the `stop_combat` method should be called.

        """
        raise NotImplementedError

    def stop_combat(self):
        """
        Stop combat. This should also do all cleanup.
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-


from django.db import migrations


def convert_channelnames(apps, schema_editor):
    ChannelDB = apps.get_model("comms", "ChannelDB")
    for chan in ChannelDB.objects.filter(db_key="MUDinfo"):
        # remove the old MUDinfo default channel
        chan.delete()
    for chan in ChannelDB.objects.filter(db_key__iexact="MUDconnections"):
        # change the old mudconnections to MudInfo instead
        chan.db_key = "MudInfo"
        chan.save()


class Migration(migrations.Migration):

    dependencies = [("comms", "0004_auto_20150118_1631")]

    operations = [migrations.RunPython(convert_channelnames)]



"""
Rerouting admin frontpage to evennia version.

These patterns are all under the admin/* namespace.

"""
from django.conf import settings
from django.contrib import admin
from django.urls import include, path

from . import frontpage

urlpatterns = [
    # Django original admin page. Make this URL always available, whether
    # we've chosen to use Evennia's custom admin or not.
    path("django/", frontpage.admin_wrapper, name="django_admin"),
    # Admin docs
    path("doc/", include("django.contrib.admindocs.urls")),
]

if settings.EVENNIA_ADMIN:
    urlpatterns += [
        # Our override for the admin.
        path("", frontpage.evennia_admin, name="evennia_admin"),
        # Makes sure that other admin pages get loaded.
        path("", admin.site.urls),
    ]
else:
    # Just include the normal Django admin.
    urlpatterns += [path("", admin.site.urls)]



"""
Unit tests for all sorts of inline text-tag parsing, like ANSI, html conversion, inlinefuncs etc

"""
import re

from django.test import TestCase, override_settings

from evennia.utils import funcparser
from evennia.utils.ansi import ANSIString
from evennia.utils.text2html import TextToHTMLparser


class ANSIStringTestCase(TestCase):
    def checker(self, ansi, raw, clean):
        """
        Verifies the raw and clean strings of an ANSIString match expected
        output.
        """
        self.assertEqual(str(ansi.clean()), clean)
        self.assertEqual(str(ansi.raw()), raw)

    def table_check(self, ansi, char, code):
        """
        Verifies the indexes in an ANSIString match what they should.
        """
        self.assertEqual(ansi._char_indexes, char)
        self.assertEqual(ansi._code_indexes, code)

    def test_instance(self):
        """
        Make sure the ANSIString is always constructed correctly.
        """
        clean = "This isA|r testTest"
        encoded = "\x1b[1m\x1b[32mThis is\x1b[1m\x1b[31mA|r test\x1b[0mTest\x1b[0m"
        target = ANSIString(r"|gThis is|rA||r test|nTest|n")
        char_table = [9, 10, 11, 12, 13, 14, 15, 25, 26, 27, 28, 29, 30, 31, 32, 37, 38, 39, 40]
        code_table = [
            0,
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            16,
            17,
            18,
            19,
            20,
            21,
            22,
            23,
            24,
            33,
            34,
            35,
            36,
            41,
            42,
            43,
            44,
        ]
        self.checker(target, encoded, clean)
        self.table_check(target, char_table, code_table)
        self.checker(ANSIString(target), encoded, clean)
        self.table_check(ANSIString(target), char_table, code_table)
        self.checker(ANSIString(encoded, decoded=True), encoded, clean)
        self.table_check(ANSIString(encoded, decoded=True), char_table, code_table)
        self.checker(ANSIString("Test"), "Test", "Test")
        self.table_check(ANSIString("Test"), [0, 1, 2, 3], [])
        self.checker(ANSIString(""), "", "")

    def test_slice(self):
        """
        Verifies that slicing an ANSIString results in expected color code
        distribution.
        """
        target = ANSIString(r"|gTest|rTest|n")
        result = target[:3]
        self.checker(result, "\x1b[1m\x1b[32mTes", "Tes")
        result = target[:4]
        self.checker(result, "\x1b[1m\x1b[32mTest\x1b[1m\x1b[31m", "Test")
        result = target[:]
        self.checker(result, "\x1b[1m\x1b[32mTest\x1b[1m\x1b[31mTest\x1b[0m", "TestTest")
        result = target[:-1]
        self.checker(result, "\x1b[1m\x1b[32mTest\x1b[1m\x1b[31mTes", "TestTes")
        result = target[0:0]
        self.checker(result, "", "")

    def test_split(self):
        """
        Verifies that re.split and .split behave similarly and that color
        codes end up where they should.
        """
        target = ANSIString("|gThis is |nA split string|g")
        first = ("\x1b[1m\x1b[32mThis is \x1b[0m", "This is ")
        second = ("\x1b[1m\x1b[32m\x1b[0m split string\x1b[1m\x1b[32m", " split string")
        re_split = re.split("A", target)
        normal_split = target.split("A")
        self.assertEqual(re_split, normal_split)
        self.assertEqual(len(normal_split), 2)
        self.checker(normal_split[0], *first)
        self.checker(normal_split[1], *second)

    def test_join(self):
        """
        Verify that joining a set of ANSIStrings works.
        """
        # This isn't the desired behavior, but the expected one. Python
        # concatenates the in-memory representation with the built-in string's
        # join.
        l = [ANSIString("|gTest|r") for _ in range(0, 3)]
        # Force the generator to be evaluated.
        result = "".join(l)
        self.assertEqual(str(result), "TestTestTest")
        result = ANSIString("").join(l)
        self.checker(
            result,
            "\x1b[1m\x1b[32mTest\x1b[1m\x1b[31m\x1b[1m\x1b"
            "[32mTest\x1b[1m\x1b[31m\x1b[1m\x1b[32mTest"
            "\x1b[1m\x1b[31m",
            "TestTestTest",
        )

    def test_len(self):
        """
        Make sure that length reporting on ANSIStrings does not include
        ANSI codes.
        """
        self.assertEqual(len(ANSIString("|gTest|n")), 4)

    def test_capitalize(self):
        """
        Make sure that capitalization works. This is the simplest of the
        _transform functions.
        """
        target = ANSIString("|gtest|n")
        result = "\x1b[1m\x1b[32mTest\x1b[0m"
        self.checker(target.capitalize(), result, "Test")

    def test_mxp_agnostic(self):
        """
        Make sure MXP tags are not treated like ANSI codes, but normal text.
        """
        mxp1 = "|lclook|ltat|le"
        mxp2 = "Start to |lclook here|ltclick somewhere here|le first"
        mxp3 = "Check out |luhttps://www.example.com|ltmy website|le!"
        self.assertEqual(15, len(ANSIString(mxp1)))
        self.assertEqual(53, len(ANSIString(mxp2)))
        self.assertEqual(53, len(ANSIString(mxp3)))
        # These would indicate an issue with the tables.
        self.assertEqual(len(ANSIString(mxp1)), len(ANSIString(mxp1).split("\n")[0]))
        self.assertEqual(len(ANSIString(mxp2)), len(ANSIString(mxp2).split("\n")[0]))
        self.assertEqual(len(ANSIString(mxp3)), len(ANSIString(mxp3).split("\n")[0]))
        self.assertEqual(mxp1, ANSIString(mxp1))
        self.assertEqual(mxp2, str(ANSIString(mxp2)))
        self.assertEqual(mxp3, str(ANSIString(mxp3)))

    def test_add(self):
        """
        Verify concatenation works correctly.
        """
        a = ANSIString("|gTest")
        b = ANSIString("|cString|n")
        c = a + b
        result = "\x1b[1m\x1b[32mTest\x1b[1m\x1b[36mString\x1b[0m"
        self.checker(c, result, "TestString")
        char_table = [9, 10, 11, 12, 22, 23, 24, 25, 26, 27]
        code_table = [0, 1, 2, 3, 4, 5, 6, 7, 8, 13, 14, 15, 16, 17, 18, 19, 20, 21, 28, 29, 30, 31]
        self.table_check(c, char_table, code_table)

    def test_strip(self):
        """
        Test the ansi-aware .strip() methods
        """
        a = ANSIString("   |r   Test of stuff |b with spaces   |n   ")
        b = ANSIString("|r|b")
        self.assertEqual(a.strip(), ANSIString("|rTest of stuff |b with spaces|n"))
        self.assertEqual(a.lstrip(), ANSIString("|rTest of stuff |b with spaces   |n   "))
        self.assertEqual(a.rstrip(), ANSIString("   |r   Test of stuff |b with spaces|n"))
        self.assertEqual(b.strip(), b)

    def test_regex_search(self):
        """
        Test regex-search in ANSIString - the found position should ignore any ansi-markers
        """
        string = ANSIString(" |r|[b  Test ")
        match = re.search(r"Test", string)
        self.assertTrue(match)
        self.assertEqual(match.span(), (3, 7))

    def test_regex_replace(self):
        """
        Inserting text into an ansistring at an index position should ignore
        the ansi markers but not remove them!

        """
        string = ANSIString("A |rTest|n string")
        match = re.search(r"Test", string)
        ix1, ix2 = match.span()
        self.assertEqual((ix1, ix2), (2, 6))

        result = string[:ix1] + "Replacement" + string[ix2:]
        expected = ANSIString("A |rReplacement|n string")

        self.assertEqual(expected, result)

    def test_slice_insert(self):
        """
        Inserting a slice should not remove ansi markup (issue #2205)
        """
        string = ANSIString("|rTest|n")
        split_string = string[:0] + "Test" + string[4:]
        self.assertEqual(string.raw(), split_string.raw())

    def test_slice_insert_longer(self):
        """
        The ANSIString replays the color code before the split in order to
        produce a *visually* identical result. The result is a longer string in
        raw characters, but one which correctly represents the color output.
        """
        string = ANSIString("A bigger |rTest|n of things |bwith more color|n")
        # from evennia import set_trace;set_trace()
        split_string = string[:9] + "Test" + string[13:]
        self.assertEqual(
            repr(
                (
                    ANSIString("A bigger ")
                    + ANSIString("|rTest")  # note that the |r|n is replayed together on next line
                    + ANSIString("|r|n of things |bwith more color|n")
                ).raw()
            ),
            repr(split_string.raw()),
        )

    def test_slice_full(self):
        string = ANSIString("A bigger |rTest|n of things |bwith more color|n")
        split_string = string[:]
        self.assertEqual(string.raw(), split_string.raw())


class TestTextToHTMLparser(TestCase):
    def setUp(self):
        self.parser = TextToHTMLparser()

    def tearDown(self):
        del self.parser

    def test_url_scheme_ftp(self):
        self.assertEqual(
            self.parser.convert_urls("ftp.example.com"),
            '<a href="http://ftp.example.com" target="_blank">ftp.example.com</a>',
        )

    def test_url_scheme_www(self):
        self.assertEqual(
            self.parser.convert_urls("www.example.com"),
            '<a href="http://www.example.com" target="_blank">www.example.com</a>',
        )

    def test_url_scheme_ftpproto(self):
        self.assertEqual(
            self.parser.convert_urls("ftp://ftp.example.com"),
            '<a href="ftp://ftp.example.com" target="_blank">ftp://ftp.example.com</a>',
        )

    def test_url_scheme_http(self):
        self.assertEqual(
            self.parser.convert_urls("http://example.com"),
            '<a href="http://example.com" target="_blank">http://example.com</a>',
        )

    def test_url_scheme_https(self):
        self.assertEqual(
            self.parser.convert_urls("https://example.com"),
            '<a href="https://example.com" target="_blank">https://example.com</a>',
        )

    def test_url_chars_slash(self):
        self.assertEqual(
            self.parser.convert_urls("www.example.com/homedir"),
            '<a href="http://www.example.com/homedir" target="_blank">www.example.com/homedir</a>',
        )

    def test_url_chars_colon(self):
        self.assertEqual(
            self.parser.convert_urls("https://example.com:8000/login/"),
            '<a href="https://example.com:8000/login/" target="_blank">'
            "https://example.com:8000/login/</a>",
        )

    def test_url_chars_querystring(self):
        self.assertEqual(
            self.parser.convert_urls("https://example.com/submitform?field1=val1+val3&field2=val2"),
            '<a href="https://example.com/submitform?field1=val1+val3&field2=val2" target="_blank">'
            "https://example.com/submitform?field1=val1+val3&field2=val2</a>",
        )

    def test_url_chars_anchor(self):
        self.assertEqual(
            self.parser.convert_urls("http://www.example.com/menu#section_1"),
            '<a href="http://www.example.com/menu#section_1" target="_blank">'
            "http://www.example.com/menu#section_1</a>",
        )

    def test_url_chars_exclam(self):
        self.assertEqual(
            self.parser.convert_urls(
                "https://groups.google.com/forum/" "?fromgroups#!categories/evennia/ainneve"
            ),
            '<a href="https://groups.google.com/forum/?fromgroups#!categories/evennia/ainneve"'
            ' target="_blank">https://groups.google.com/forum/?fromgroups#!categories/evennia/ainneve</a>',
        )

    def test_url_edge_following_period_eol(self):
        self.assertEqual(
            self.parser.convert_urls("www.example.com."),
            '<a href="http://www.example.com" target="_blank">www.example.com</a>.',
        )

    def test_url_edge_following_period(self):
        self.assertEqual(
            self.parser.convert_urls("see www.example.com. "),
            'see <a href="http://www.example.com" target="_blank">www.example.com</a>. ',
        )

    def test_url_edge_brackets(self):
        self.assertEqual(
            self.parser.convert_urls("[http://example.com/]"),
            '[<a href="http://example.com/" target="_blank">http://example.com/</a>]',
        )

    def test_url_edge_multiline(self):
        self.assertEqual(
            self.parser.convert_urls("  * http://example.com/info\n  * bullet"),
            '  * <a href="http://example.com/info" target="_blank">'
            "http://example.com/info</a>\n  * bullet",
        )

    def test_url_edge_following_htmlentity(self):
        self.assertEqual(
            self.parser.convert_urls("http://example.com/info&lt;span&gt;"),
            '<a href="http://example.com/info" target="_blank">http://example.com/info</a>&lt;span&gt;',
        )

    def test_url_edge_surrounded_spans(self):
        self.assertEqual(
            self.parser.convert_urls('</span>http://example.com/<span class="red">'),
            '</span><a href="http://example.com/" target="_blank">'
            'http://example.com/</a><span class="red">',
        )

    def test_non_url_with_www(self):
        self.assertEqual(
            self.parser.convert_urls("Awwww.this should not be highlighted"),
            "Awwww.this should not be highlighted",
        )

    def test_invalid_www_url(self):
        self.assertEqual(self.parser.convert_urls("www.t"), "www.t")



"""
This module implements a simple mobile object with
a very rudimentary AI as well as an aggressive enemy
object based on that mobile class.

"""

import random

from evennia import TICKER_HANDLER, CmdSet, Command, logger, search_object

from . import objects as tut_objects


class CmdMobOnOff(Command):
    """
    Activates/deactivates Mob

    Usage:
        mobon <mob>
        moboff <mob>

    This turns the mob from active (alive) mode
    to inactive (dead) mode. It is used during
    building to  activate the mob once it's
    prepared.
    """

    key = "mobon"
    aliases = "moboff"
    locks = "cmd:superuser()"

    def func(self):
        """
        Uses the mob's set_alive/set_dead methods
        to turn on/off the mob."
        """
        if not self.args:
            self.caller.msg("Usage: mobon||moboff <mob>")
            return
        mob = self.caller.search(self.args)
        if not mob:
            return
        if self.cmdstring == "mobon":
            mob.set_alive()
        else:
            mob.set_dead()


class MobCmdSet(CmdSet):
    """
    Holds the admin command controlling the mob
    """

    def at_cmdset_creation(self):
        self.add(CmdMobOnOff())


class Mob(tut_objects.TutorialObject):
    """
    This is a state-machine AI mobile. It has several states which are
    controlled from setting various Attributes. All default to True:

        patrolling: if set, the mob will move randomly
            from room to room, but preferring to not return
            the way it came. If unset, the mob will remain
            stationary (idling) until attacked.
        aggressive: if set, will attack Characters in
            the same room using whatever Weapon it
            carries (see tutorial_world.objects.TutorialWeapon).
            if unset, the mob will never engage in combat
            no matter what.
        hunting: if set, the mob will pursue enemies trying
            to flee from it, so it can enter combat. If unset,
            it will return to patrolling/idling if fled from.
        immortal: If set, the mob cannot take any damage.
        irregular_echoes: list of strings the mob generates at irregular intervals.
        desc_alive: the physical description while alive
        desc_dead: the physical descripion while dead
        send_defeated_to: unique key/alias for location to send defeated enemies to
        defeat_msg: message to echo to defeated opponent
        defeat_msg_room: message to echo to room. Accepts %s as the name of the defeated.
        hit_msg: message to echo when this mob is hit. Accepts %s for the mob's key.
        weapon_ineffective_msg: message to echo for useless attacks
        death_msg: message to echo to room when this mob dies.
        patrolling_pace: how many seconds per tick, when patrolling
        aggressive_pace:   -"-         attacking
        hunting_pace:      -"-         hunting
        death_pace:        -"-         returning to life when dead

        field 'home' - the home location should set to someplace inside
           the patrolling area. The mob will use this if it should
           happen to roam into a room with no exits.

    """

    def at_init(self):
        """
        When initialized from cache (after a server reboot), set up
        the AI state.
        """
        # The AI state machine (not persistent).
        self.ndb.is_patrolling = self.db.patrolling and not self.db.is_dead
        self.ndb.is_attacking = False
        self.ndb.is_hunting = False
        self.ndb.is_immortal = self.db.immortal or self.db.is_dead

    def at_object_creation(self):
        """
        Called the first time the object is created.
        We set up the base properties and flags here.
        """
        self.cmdset.add(MobCmdSet, persistent=True)
        # Main AI flags. We start in dead mode so we don't have to
        # chase the mob around when building.
        self.db.patrolling = True
        self.db.aggressive = True
        self.db.immortal = False
        # db-store if it is dead or not
        self.db.is_dead = True
        # specifies how much damage we divide away from non-magic weapons
        self.db.damage_resistance = 100.0
        # pace (number of seconds between ticks) for
        # the respective modes.
        self.db.patrolling_pace = 6
        self.db.aggressive_pace = 2
        self.db.hunting_pace = 1
        self.db.death_pace = 100  # stay dead for 100 seconds

        # we store the call to the tickerhandler
        # so we can easily deactivate the last
        # ticker subscription when we switch.
        # since we will use the same idstring
        # throughout we only need to save the
        # previous interval we used.
        self.db.last_ticker_interval = None

        # store two separate descriptions, one for alive and
        # one for dead (corpse)
        self.db.desc_alive = "This is a moving object."
        self.db.desc_dead = "A dead body."

        # health stats
        self.db.full_health = 20
        self.db.health = 20

        # when this mob defeats someone, we move the character off to
        # some other place (Dark Cell in the tutorial).
        self.db.send_defeated_to = "dark cell"
        # text to echo to the defeated foe.
        self.db.defeat_msg = "You fall to the ground."
        self.db.defeat_msg_room = "%s falls to the ground."
        self.db.weapon_ineffective_msg = (
            "Your weapon just passes through your enemy, causing almost no effect!"
        )

        self.db.death_msg = "After the last hit %s evaporates." % self.key
        self.db.hit_msg = "%s wails, shudders and writhes." % self.key
        self.db.irregular_msgs = ["the enemy looks about.", "the enemy changes stance."]

        self.db.tutorial_info = "This is an object with simple state AI, using a ticker to move."

    def _set_ticker(self, interval, hook_key, stop=False):
        """
        Set how often the given hook key should
        be "ticked".

        Args:
            interval (int or None): The number of seconds
                between ticks
            hook_key (str or None): The name of the method
                (on this mob) to call every interval
                seconds.
            stop (bool, optional): Just stop the
                last ticker without starting a new one.
                With this set, the interval and hook_key
                arguments are unused.

        In order to only have one ticker
        running at a time, we make sure to store the
        previous ticker subscription so that we can
        easily find and stop it before setting a
        new one. The tickerhandler is persistent so
        we need to remember this across reloads.

        """
        idstring = "tutorial_mob"  # this doesn't change
        last_interval = self.db.last_ticker_interval
        last_hook_key = self.db.last_hook_key
        if last_interval and last_hook_key:
            # we have a previous subscription, kill this first.
            TICKER_HANDLER.remove(
                interval=last_interval, callback=getattr(self, last_hook_key), idstring=idstring
            )
        self.db.last_ticker_interval = interval
        self.db.last_hook_key = hook_key
        if not stop:
            # set the new ticker
            TICKER_HANDLER.add(
                interval=interval, callback=getattr(self, hook_key), idstring=idstring
            )

    def _find_target(self, location):
        """
        Scan the given location for suitable targets (this is defined
        as Characters) to attack.  Will ignore superusers.

        Args:
            location (Object): the room to scan.

        Returns:
            The first suitable target found.

        """
        targets = [
            obj
            for obj in location.contents_get(exclude=self)
            if obj.has_account and not obj.is_superuser
        ]
        return targets[0] if targets else None

    def set_alive(self, *args, **kwargs):
        """
        Set the mob to "alive" mode. This effectively
        resurrects it from the dead state.
        """
        self.db.health = self.db.full_health
        self.db.is_dead = False
        self.db.desc = self.db.desc_alive
        self.ndb.is_immortal = self.db.immortal
        self.ndb.is_patrolling = self.db.patrolling
        if not self.location:
            self.move_to(self.home)
        if self.db.patrolling:
            self.start_patrolling()

    def set_dead(self):
        """
        Set the mob to "dead" mode. This turns it off
        and makes sure it can take no more damage.
        It also starts a ticker for when it will return.
        """
        self.db.is_dead = True
        self.location = None
        self.ndb.is_patrolling = False
        self.ndb.is_attacking = False
        self.ndb.is_hunting = False
        self.ndb.is_immortal = True
        # we shall return after some time
        self._set_ticker(self.db.death_pace, "set_alive")

    def start_idle(self):
        """
        Starts just standing around. This will kill
        the ticker and do nothing more.
        """
        self._set_ticker(None, None, stop=True)

    def start_patrolling(self):
        """
        Start the patrolling state by
        registering us with the ticker-handler
        at a leasurely pace.
        """
        if not self.db.patrolling:
            self.start_idle()
            return
        self._set_ticker(self.db.patrolling_pace, "do_patrol")
        self.ndb.is_patrolling = True
        self.ndb.is_hunting = False
        self.ndb.is_attacking = False
        # for the tutorial, we also heal the mob in this mode
        self.db.health = self.db.full_health

    def start_hunting(self):
        """
        Start the hunting state
        """
        if not self.db.hunting:
            self.start_patrolling()
            return
        self._set_ticker(self.db.hunting_pace, "do_hunt")
        self.ndb.is_patrolling = False
        self.ndb.is_hunting = True
        self.ndb.is_attacking = False

    def start_attacking(self):
        """
        Start the attacking state
        """
        if not self.db.aggressive:
            self.start_hunting()
            return
        self._set_ticker(self.db.aggressive_pace, "do_attack")
        self.ndb.is_patrolling = False
        self.ndb.is_hunting = False
        self.ndb.is_attacking = True

    def do_patrol(self, *args, **kwargs):
        """
        Called repeatedly during patrolling mode.  In this mode, the
        mob scans its surroundings and randomly chooses a viable exit.
        One should lock exits with the traverse:has_account() lock in
        order to block the mob from moving outside its area while
        allowing account-controlled characters to move normally.
        """
        if random.random() < 0.01 and self.db.irregular_msgs:
            self.location.msg_contents(random.choice(self.db.irregular_msgs))
        if self.db.aggressive:
            # first check if there are any targets in the room.
            target = self._find_target(self.location)
            if target:
                self.start_attacking()
                return
        # no target found, look for an exit.
        exits = [exi for exi in self.location.exits if exi.access(self, "traverse")]
        if exits:
            # randomly pick an exit
            exit = random.choice(exits)
            # move there.
            self.move_to(exit.destination)
        else:
            # no exits! teleport to home to get away.
            self.move_to(self.home)

    def do_hunting(self, *args, **kwargs):
        """
        Called regularly when in hunting mode. In hunting mode the mob
        scans adjacent rooms for enemies and moves towards them to
        attack if possible.
        """
        if random.random() < 0.01 and self.db.irregular_msgs:
            self.location.msg_contents(random.choice(self.db.irregular_msgs))
        if self.db.aggressive:
            # first check if there are any targets in the room.
            target = self._find_target(self.location)
            if target:
                self.start_attacking()
                return
        # no targets found, scan surrounding rooms
        exits = [exi for exi in self.location.exits if exi.access(self, "traverse")]
        if exits:
            # scan the exits destination for targets
            for exit in exits:
                target = self._find_target(exit.destination)
                if target:
                    # a target found. Move there.
                    self.move_to(exit.destination)
                    return
            # if we get to this point we lost our
            # prey. Resume patrolling.
            self.start_patrolling()
        else:
            # no exits! teleport to home to get away.
            self.move_to(self.home)

    def do_attack(self, *args, **kwargs):
        """
        Called regularly when in attacking mode. In attacking mode
        the mob will bring its weapons to bear on any targets
        in the room.
        """
        if random.random() < 0.01 and self.db.irregular_msgs:
            self.location.msg_contents(random.choice(self.db.irregular_msgs))
        # first make sure we have a target
        target = self._find_target(self.location)
        if not target:
            # no target, start looking for one
            self.start_hunting()
            return

        # we use the same attack commands as defined in
        # tutorial_world.objects.TutorialWeapon, assuming that
        # the mob is given a Weapon to attack with.
        attack_cmd = random.choice(("thrust", "pierce", "stab", "slash", "chop"))
        self.execute_cmd("%s %s" % (attack_cmd, target))

        # analyze the current state
        if target.db.health <= 0:
            # we reduced the target to <= 0 health. Move them to the
            # defeated room
            target.msg(self.db.defeat_msg)
            self.location.msg_contents(self.db.defeat_msg_room % target.key, exclude=target)
            send_defeated_to = search_object(self.db.send_defeated_to)
            if send_defeated_to:
                target.move_to(send_defeated_to[0], quiet=True)
            else:
                logger.log_err(
                    "Mob: mob.db.send_defeated_to not found: %s" % self.db.send_defeated_to
                )

    # response methods - called by other objects

    def at_hit(self, weapon, attacker, damage):
        """
        Someone landed a hit on us. Check our status
        and start attacking if not already doing so.
        """
        if self.db.health is None:
            # health not set - this can't be damaged.
            attacker.msg(self.db.weapon_ineffective_msg)
            return

        if not self.ndb.is_immortal:
            if not weapon.db.magic:
                # not a magic weapon - divide away magic resistance
                damage /= self.db.damage_resistance
                attacker.msg(self.db.weapon_ineffective_msg)
            else:
                self.location.msg_contents(self.db.hit_msg)
            self.db.health -= damage

        # analyze the result
        if self.db.health <= 0:
            # we are dead!
            attacker.msg(self.db.death_msg)
            self.set_dead()
        else:
            # still alive, start attack if not already attacking
            if self.db.aggressive and not self.ndb.is_attacking:
                self.start_attacking()

    def at_new_arrival(self, new_character):
        """
        This is triggered whenever a new character enters the room.
        This is called by the TutorialRoom the mob stands in and
        allows it to be aware of changes immediately without needing
        to poll for them all the time. For example, the mob can react
        right away, also when patrolling on a very slow ticker.
        """
        # the room actually already checked all we need, so
        # we know it is a valid target.
        if self.db.aggressive and not self.ndb.is_attacking:
            self.start_attacking()



# Evadventure (Turnbased) combat demo - using a batch-code file.
#
# Sets up a combat area for testing turnbased combat.
#
# First add mygame/server/conf/settings.py:
#
#    BASE_BATCHPROCESS_PATHS += ["evadventure.batchscripts"]
#
# Run from in-game as `batchcode turnbased_combat_demo`
#

# HEADER

from evennia import DefaultExit, create_object, search_object
from evennia.contrib.tutorials.evadventure.characters import EvAdventureCharacter
from evennia.contrib.tutorials.evadventure.combat_turnbased import TurnCombatCmdSet
from evennia.contrib.tutorials.evadventure.npcs import EvAdventureNPC
from evennia.contrib.tutorials.evadventure.rooms import EvAdventureRoom

# CODE

# Make the player an EvAdventureCharacter
player = caller  # caller is injected by the batchcode runner, it's the one running this script
player.swap_typeclass(EvAdventureCharacter)

# add the Turnbased cmdset
player.cmdset.add(TurnCombatCmdSet, persistent=True)

# create a weapon and an item to use
create_object(
    "contrib.tutorials.evadventure.objects.EvAdventureWeapon",
    key="Sword",
    location=player,
    attributes=[("desc", "A sword.")],
)

create_object(
    "contrib.tutorials.evadventure.objects.EvAdventureConsumable",
    key="Potion",
    location=player,
    attributes=[("desc", "A potion.")],
)

# start from limbo
limbo = search_object("#2")[0]

arena = create_object(EvAdventureRoom, key="Arena", attributes=[("desc", "A large arena.")])

# Create the exits
arena_exit = create_object(DefaultExit, key="Arena", location=limbo, destination=arena)
back_exit = create_object(DefaultExit, key="Back", location=arena, destination=limbo)

# create the NPC dummy
create_object(
    EvAdventureNPC,
    key="Dummy",
    location=arena,
    attributes=[("desc", "A training dummy."), ("hp", 1000), ("hp_max", 1000)],
)



"""
Health Bar

Contrib - Tim Ashley Jenkins 2017

The function provided in this module lets you easily display visual
bars or meters - "health bar" is merely the most obvious use for this,
though these bars are highly customizable and can be used for any sort
of appropriate data besides player health.

Today's players may be more used to seeing statistics like health,
stamina, magic, and etc. displayed as bars rather than bare numerical
values, so using this module to present this data this way may make it
more accessible. Keep in mind, however, that players may also be using
a screen reader to connect to your game, which will not be able to
represent the colors of the bar in any way. By default, the values
represented are rendered as text inside the bar which can be read by
screen readers.

The health bar will account for current values above the maximum or
below 0, rendering them as a completely full or empty bar with the
values displayed within.

No installation, just import and use `display_meter` from this
module:

```python
    from evennia.contrib.rpg.health_bar import display_meter

    health_bar = display_meter(23, 100)
    caller.msg(prompt=health_bar)
```

"""


def display_meter(
    cur_value,
    max_value,
    length=30,
    fill_color=["R", "Y", "G"],
    empty_color="B",
    text_color="w",
    align="left",
    pre_text="",
    post_text="",
    show_values=True,
):
    """
    Represents a current and maximum value given as a "bar" rendered with
    ANSI or xterm256 background colors.

    Args:
        cur_value (int): Current value to display
        max_value (int): Maximum value to display

    Keyword Arguments:
        length (int): Length of meter returned, in characters
        fill_color (list): List of color codes for the full portion
            of the bar, sans any sort of prefix - both ANSI and xterm256
            colors are usable. When the bar is empty, colors toward the
            start of the list will be chosen - when the bar is full, colors
            towards the end are picked. You can adjust the 'weights' of
            the changing colors by adding multiple entries of the same
            color - for example, if you only want the bar to change when
            it's close to empty, you could supply ['R','Y','G','G','G']
        empty_color (str): Color code for the empty portion of the bar.
        text_color (str): Color code for text inside the bar.
        align (str): "left", "right", or "center" - alignment of text in the bar
        pre_text (str): Text to put before the numbers in the bar
        post_text (str): Text to put after the numbers in the bar
        show_values (bool): If true, shows the numerical values represented by
            the bar. It's highly recommended you keep this on, especially if
            there's no info given in pre_text or post_text, as players on screen
            readers will be unable to read the graphical aspect of the bar.

    Returns:
        str: The display bar to show.

    """
    # Start by building the base string.
    num_text = ""
    if show_values:
        num_text = "%i / %i" % (cur_value, max_value)
    bar_base_str = pre_text + num_text + post_text
    # Cut down the length of the base string if needed
    if len(bar_base_str) > length:
        bar_base_str = bar_base_str[:length]
    # Pad and align the bar base string
    if align == "right":
        bar_base_str = bar_base_str.rjust(length, " ")
    elif align == "center":
        bar_base_str = bar_base_str.center(length, " ")
    else:
        bar_base_str = bar_base_str.ljust(length, " ")

    if max_value < 1:  # Prevent divide by zero
        max_value = 1
    if cur_value < 0:  # Prevent weirdly formatted 'negative bars'
        cur_value = 0
    if cur_value > max_value:  # Display overfull bars correctly
        cur_value = max_value

    # Now it's time to determine where to put the color codes.
    percent_full = float(cur_value) / float(max_value)
    split_index = round(float(length) * percent_full)
    # Determine point at which to split the bar
    split_index = int(split_index)

    # Separate the bar string into full and empty portions
    full_portion = bar_base_str[:split_index]
    empty_portion = bar_base_str[split_index:]

    # Pick which fill color to use based on how full the bar is
    fillcolor_index = float(len(fill_color)) * percent_full
    fillcolor_index = max(0, int(round(fillcolor_index)) - 1)
    fillcolor_code = "|[" + fill_color[fillcolor_index]

    # Make color codes for empty bar portion and text_color
    emptycolor_code = "|[" + empty_color
    textcolor_code = "|" + text_color

    # Assemble the final bar
    final_bar = (
        fillcolor_code
        + textcolor_code
        + full_portion
        + "|n"
        + emptycolor_code
        + textcolor_code
        + empty_portion
        + "|n"
    )

    return final_bar



"""
TutorialMirror

A simple mirror object to experiment with.

"""

from evennia import DefaultObject, logger
from evennia.utils import is_iter, make_iter


class TutorialMirror(DefaultObject):
    """
    A simple mirror object that
    - echoes back the description of the object looking at it
    - echoes back whatever is being sent to its .msg - to the
      sender, if given, otherwise to the location of the mirror.

    """

    def return_appearance(self, looker, **kwargs):
        """
        This formats the description of this object. Called by the 'look' command.

        Args:
            looker (Object): Object doing the looking.
            **kwargs (dict): Arbitrary, optional arguments for users
                overriding the call (unused by default).
        """

        if isinstance(looker, self.__class__):
            # avoid infinite recursion by having two mirrors look at each other
            return "The image of yourself stretches into infinity."
        return f"{self.key} shows your reflection:\n{looker.db.desc}"

    def msg(self, text=None, from_obj=None, **kwargs):
        """
        Simply override .msg to echo back to the messenger or to the current
        location.

        Args:
            text (str or tuple, optional): The message to send. This
                is treated internally like any send-command, so its
                value can be a tuple if sending multiple arguments to
                the `text` oob command.
            from_obj (obj or iterable)
                given, at_msg_send will be called. This value will be
                passed on to the protocol. If iterable, will execute hook
                on all entities in it.
        """
        if not text:
            text = "<silence>"
        text = text[0] if is_iter(text) else text
        if from_obj:
            for obj in make_iter(from_obj):
                obj.msg(f'{self.key} echoes back to you:\n"{text}".')
        elif self.location:
            self.location.msg_contents(f'{self.key} echoes back:\n"{text}".', exclude=[self])
        else:
            # no from_obj and no location, just log
            logger.log_msg(f"{self.key}.msg was called without from_obj and .location is None.")



"""
Django ID mapper

Modified for Evennia by making sure that no model references
leave caching unexpectedly (no use of WeakRefs).

Also adds `cache_size()` for monitoring the size of the cache.
"""

import gc
import os
import threading
import time
from weakref import WeakValueDictionary

from django.core.exceptions import FieldError, ObjectDoesNotExist
from django.db.models.base import Model, ModelBase
from django.db.models.signals import post_migrate, post_save, pre_delete
from django.db.transaction import atomic
from django.db.utils import DatabaseError
from twisted.internet.reactor import callFromThread

from evennia.utils import logger
from evennia.utils.utils import dbref, get_evennia_pids, to_str

from .manager import SharedMemoryManager

AUTO_FLUSH_MIN_INTERVAL = 60.0 * 5  # at least 5 mins between cache flushes

_GA = object.__getattribute__
_SA = object.__setattr__
_DA = object.__delattr__
_MONITOR_HANDLER = None

# References to db-updated objects are stored here so the
# main process can be informed to re-cache itself.
PROC_MODIFIED_COUNT = 0
PROC_MODIFIED_OBJS = WeakValueDictionary()

# get info about the current process and thread; determine if our
# current pid is different from the server PID (i.e.  # if we are in a
# subprocess or not)
_SELF_PID = os.getpid()
_SERVER_PID, _PORTAL_PID = get_evennia_pids()
_IS_SUBPROCESS = (_SERVER_PID and _PORTAL_PID) and _SELF_PID not in (_SERVER_PID, _PORTAL_PID)
_IS_MAIN_THREAD = threading.current_thread().name == "MainThread"


class SharedMemoryModelBase(ModelBase):
    # CL: upstream had a __new__ method that skipped ModelBase's __new__ if
    # SharedMemoryModelBase was not in the model class's ancestors. It's not
    # clear what was the intended purpose, but skipping ModelBase.__new__
    # broke things; in particular, default manager inheritance.

    def __call__(cls, *args, **kwargs):
        """
        this method will either create an instance (by calling the default implementation)
        or try to retrieve one from the class-wide cache by inferring the pk value from
        `args` and `kwargs`. If instance caching is enabled for this class, the cache is
        populated whenever possible (ie when it is possible to infer the pk value).

        """

        def new_instance():
            return super(SharedMemoryModelBase, cls).__call__(*args, **kwargs)

        instance_key = cls._get_cache_key(args, kwargs)
        # depending on the arguments, we might not be able to infer the PK, so in that case we
        # create a new instance
        if instance_key is None:
            return new_instance()
        cached_instance = cls.get_cached_instance(instance_key)
        if cached_instance is None:
            cached_instance = new_instance()
            cls.cache_instance(cached_instance, new=True)
        return cached_instance

    def _prepare(cls):
        """
        Prepare the cache, making sure that proxies of the same db base
        share the same cache.

        """
        # the dbmodel is either the proxy base or ourselves
        dbmodel = cls._meta.concrete_model if cls._meta.proxy else cls
        cls.__dbclass__ = dbmodel
        if not hasattr(dbmodel, "__instance_cache__"):
            # we store __instance_cache__ only on the dbmodel base
            dbmodel.__instance_cache__ = {}
        super()._prepare()

    def __new__(cls, name, bases, attrs):
        """
        Field shortcut creation:

        Takes field names `db_*` and creates property wrappers named
        without the `db_` prefix. So db_key -> key

        This wrapper happens on the class level, so there is no
        overhead when creating objects.  If a class already has a
        wrapper of the given name, the automatic creation is skipped.

        Notes:
            Remember to document this auto-wrapping in the class
            header, this could seem very much like magic to the user
            otherwise.
        """

        attrs["typename"] = cls.__name__
        attrs["path"] = "%s.%s" % (attrs["__module__"], name)
        attrs["_is_deleted"] = False

        # set up the typeclass handling only if a variable _is_typeclass is set on the class
        def create_wrapper(cls, fieldname, wrappername, editable=True, foreignkey=False):
            "Helper method to create property wrappers with unique names (must be in separate call)"

            def _get(cls, fname):
                "Wrapper for getting database field"
                if _GA(cls, "_is_deleted"):
                    raise ObjectDoesNotExist(
                        "Cannot access %s: Hosting object was already deleted." % fname
                    )
                return _GA(cls, fieldname)

            def _get_foreign(cls, fname):
                "Wrapper for returning foreignkey fields"
                if _GA(cls, "_is_deleted"):
                    raise ObjectDoesNotExist(
                        "Cannot access %s: Hosting object was already deleted." % fname
                    )
                return _GA(cls, fieldname)

            def _set_nonedit(cls, fname, value):
                "Wrapper for blocking editing of field"
                raise FieldError("Field %s cannot be edited." % fname)

            def _set(cls, fname, value):
                "Wrapper for setting database field"
                if _GA(cls, "_is_deleted"):
                    raise ObjectDoesNotExist(
                        "Cannot set %s to %s: Hosting object was already deleted!" % (fname, value)
                    )
                _SA(cls, fname, value)
                # only use explicit update_fields in save if we actually have a
                # primary key assigned already (won't be set when first creating object)
                update_fields = (
                    [fname] if _GA(cls, "_get_pk_val")(_GA(cls, "_meta")) is not None else None
                )
                _GA(cls, "save")(update_fields=update_fields)

            def _set_foreign(cls, fname, value):
                "Setter only used on foreign key relations, allows setting with #dbref"
                if _GA(cls, "_is_deleted"):
                    raise ObjectDoesNotExist(
                        "Cannot set %s to %s: Hosting object was already deleted!" % (fname, value)
                    )
                if isinstance(value, (str, int)):
                    value = to_str(value)
                    if value.isdigit() or value.startswith("#"):
                        # we also allow setting using dbrefs, if so we try to load the matching
                        # object. (we assume the object is of the same type as the class holding
                        # the field, if not a custom handler must be used for that field)
                        dbid = dbref(value, reqhash=False)
                        if dbid:
                            model = _GA(cls, "_meta").get_field(fname).model
                            try:
                                value = model._default_manager.get(id=dbid)
                            except ObjectDoesNotExist:
                                # maybe it is just a name that happens to look like a dbid
                                pass
                _SA(cls, fname, value)
                # only use explicit update_fields in save if we actually have a
                # primary key assigned already (won't be set when first creating object)
                update_fields = (
                    [fname] if _GA(cls, "_get_pk_val")(_GA(cls, "_meta")) is not None else None
                )
                _GA(cls, "save")(update_fields=update_fields)

            def _del_nonedit(cls, fname):
                "wrapper for not allowing deletion"
                raise FieldError("Field %s cannot be edited." % fname)

            def _del(cls, fname):
                "Wrapper for clearing database field - sets it to None"
                _SA(cls, fname, None)
                update_fields = (
                    [fname] if _GA(cls, "_get_pk_val")(_GA(cls, "_meta")) is not None else None
                )
                _GA(cls, "save")(update_fields=update_fields)

            # wrapper factories
            if not editable:

                def fget(cls):
                    return _get(cls, fieldname)

                def fset(cls, val):
                    return _set_nonedit(cls, fieldname, val)

            elif foreignkey:

                def fget(cls):
                    return _get_foreign(cls, fieldname)

                def fset(cls, val):
                    return _set_foreign(cls, fieldname, val)

            else:

                def fget(cls):
                    return _get(cls, fieldname)

                def fset(cls, val):
                    return _set(cls, fieldname, val)

            def fdel(cls):
                return _del(cls, fieldname) if editable else _del_nonedit(cls, fieldname)

            # set docstrings for auto-doc
            fget.__doc__ = "A wrapper for getting database field `%s`." % fieldname
            fset.__doc__ = "A wrapper for setting (and saving) database field `%s`." % fieldname
            fdel.__doc__ = "A wrapper for deleting database field `%s`." % fieldname
            # assigning
            attrs[wrappername] = property(fget, fset, fdel)
            # type(cls).__setattr__(cls, wrappername, property(fget, fset, fdel))#, doc))

        # exclude some models that should not auto-create wrapper fields
        if cls.__name__ in ("ServerConfig", "TypeNick"):
            return
        # dynamically create the wrapper properties for all fields not already handled
        # (manytomanyfields are always handlers)
        for fieldname, field in (
            (fname, field)
            for fname, field in list(attrs.items())
            if fname.startswith("db_") and type(field).__name__ != "ManyToManyField"
        ):
            foreignkey = type(field).__name__ == "ForeignKey"
            wrappername = "dbid" if fieldname == "id" else fieldname.replace("db_", "", 1)
            if wrappername not in attrs:
                # makes sure not to overload manually created wrappers on the model
                create_wrapper(
                    cls, fieldname, wrappername, editable=field.editable, foreignkey=foreignkey
                )

        return super().__new__(cls, name, bases, attrs)


class SharedMemoryModel(Model, metaclass=SharedMemoryModelBase):
    """
    Base class for idmapped objects. Inherit from `this`.
    """

    objects = SharedMemoryManager()

    class Meta(object):
        abstract = True

    @classmethod
    def _get_cache_key(cls, args, kwargs):
        """
        This method is used by the caching subsystem to infer the PK
        value from the constructor arguments.  It is used to decide if
        an instance has to be built or is already in the cache.

        """
        result = None
        # Quick hack for my composites work for now.
        if hasattr(cls._meta, "pks"):
            pk = cls._meta.pks[0]
        else:
            pk = cls._meta.pk
        # get the index of the pk in the class fields. this should be calculated *once*, but isn't
        # atm
        pk_position = cls._meta.fields.index(pk)
        if len(args) > pk_position:
            # if it's in the args, we can get it easily by index
            result = args[pk_position]
        elif pk.attname in kwargs:
            # retrieve the pk value. Note that we use attname instead of name, to handle the case
            # where the pk is a a ForeignKey.
            result = kwargs[pk.attname]
        elif pk.name != pk.attname and pk.name in kwargs:
            # ok we couldn't find the value, but maybe it's a FK and we can find the corresponding
            # object instead
            result = kwargs[pk.name]

        if result is not None and isinstance(result, Model):
            # if the pk value happens to be a model instance (which can happen wich a FK), we'd
            # rather use its own pk as the key
            result = result._get_pk_val()
        return result

    @classmethod
    def get_cached_instance(cls, id):
        """
        Method to retrieve a cached instance by pk value. Returns None
        when not found (which will always be the case when caching is
        disabled for this class). Please note that the lookup will be
        done even when instance caching is disabled.

        """
        return cls.__dbclass__.__instance_cache__.get(id)

    @classmethod
    def cache_instance(cls, instance, new=False):
        """
        Method to store an instance in the cache.

        Args:
            instance (Class instance): the instance to cache.
            new (bool, optional): this is the first time this instance is
                cached (i.e. this is not an update operation like after a
                db save).

        """
        pk = instance._get_pk_val()
        if pk is not None:
            new = new or pk not in cls.__dbclass__.__instance_cache__
            cls.__dbclass__.__instance_cache__[pk] = instance
            if new:
                try:
                    # trigger the at_init hook only
                    # at first initialization
                    instance.at_init()
                except AttributeError:
                    # The at_init hook is not assigned to all entities
                    pass

    @classmethod
    def get_all_cached_instances(cls):
        """
        Return the objects so far cached by idmapper for this class.

        """
        return list(cls.__dbclass__.__instance_cache__.values())

    @classmethod
    def _flush_cached_by_key(cls, key, force=True):
        """
        Remove the cached reference.

        """
        try:
            if force or cls.at_idmapper_flush():
                del cls.__dbclass__.__instance_cache__[key]
            else:
                cls._dbclass__.__instance_cache__[key].refresh_from_db()
        except KeyError:
            # No need to remove if cache doesn't contain it already
            pass

    @classmethod
    def flush_cached_instance(cls, instance, force=True):
        """
        Method to flush an instance from the cache. The instance will
        always be flushed from the cache, since this is most likely
        called from delete(), and we want to make sure we don't cache
        dead objects.

        """
        cls._flush_cached_by_key(instance._get_pk_val(), force=force)

    # flush_cached_instance = classmethod(flush_cached_instance)

    @classmethod
    def flush_instance_cache(cls, force=False):
        """
        This will clean safe objects from the cache. Use `force`
        keyword to remove all objects, safe or not.

        """
        if force:
            cls.__dbclass__.__instance_cache__ = {}
        else:
            cls.__dbclass__.__instance_cache__ = dict(
                (key, obj)
                for key, obj in cls.__dbclass__.__instance_cache__.items()
                if not obj.at_idmapper_flush()
            )

    # flush_instance_cache = classmethod(flush_instance_cache)

    # per-instance methods

    def __eq__(self, other):
        return super().__eq__(other)

    def __hash__(self):
        # this is required to maintain hashing
        return super().__hash__()

    def at_idmapper_flush(self):
        """
        This is called when the idmapper cache is flushed and
        allows customized actions when this happens.

        Returns:
            do_flush (bool): If True, flush this object as normal. If
                False, don't flush and expect this object to handle
                the flushing on its own.
        """
        return True

    def flush_from_cache(self, force=False):
        """
        Flush this instance from the instance cache. Use
        `force` to override the result of at_idmapper_flush() for the object.

        """
        pk = self._get_pk_val()
        if pk:
            if force or self.at_idmapper_flush():
                self.__class__.__dbclass__.__instance_cache__.pop(pk, None)

    def delete(self, *args, **kwargs):
        """
        Delete the object, clearing cache.

        """
        self.flush_from_cache()
        self._is_deleted = True
        super().delete(*args, **kwargs)

    def save(self, *args, **kwargs):
        """
        Central database save operation.

        Notes:
            Arguments as per Django documentation.
            Calls `self.at_<fieldname>_postsave(new)`
            (this is a wrapper set by oobhandler:
            self._oob_at_<fieldname>_postsave())

        """
        global _MONITOR_HANDLER
        if not _MONITOR_HANDLER:
            from evennia.scripts.monitorhandler import (
                MONITOR_HANDLER as _MONITOR_HANDLER,
            )

        if _IS_SUBPROCESS:
            # we keep a store of objects modified in subprocesses so
            # we know to update their caches in the central process
            global PROC_MODIFIED_COUNT, PROC_MODIFIED_OBJS
            PROC_MODIFIED_COUNT += 1
            PROC_MODIFIED_OBJS[PROC_MODIFIED_COUNT] = self

        if _IS_MAIN_THREAD:
            # in main thread - normal operation
            try:
                with atomic():
                    super().save(*args, **kwargs)
            except DatabaseError:
                # we handle the 'update_fields did not update any rows' error that
                # may happen due to timing issues with attributes
                ufields_removed = kwargs.pop("update_fields", None)
                if ufields_removed:
                    with atomic():
                        super().save(*args, **kwargs)
                else:
                    raise
        else:
            # in another thread; make sure to save in reactor thread
            def _save_callback(cls, *args, **kwargs):
                super().save(*args, **kwargs)

            callFromThread(_save_callback, self, *args, **kwargs)

        if not self.pk:
            # this can happen if some of the startup methods immediately
            # delete the object (an example are Scripts that start and die immediately)
            return

        # update field-update hooks and eventual OOB watchers
        new = False
        if "update_fields" in kwargs and kwargs["update_fields"]:
            # get field objects from their names
            update_fields = (
                self._meta.get_field(fieldname) for fieldname in kwargs.get("update_fields")
            )
        else:
            # meta.fields are already field objects; get them all
            new = True
            update_fields = self._meta.fields
        for field in update_fields:
            fieldname = field.name
            # trigger eventual monitors
            _MONITOR_HANDLER.at_update(self, fieldname)
            # if a hook is defined it must be named exactly on this form
            hookname = "at_%s_postsave" % fieldname
            if hasattr(self, hookname) and callable(_GA(self, hookname)):
                _GA(self, hookname)(new)

        #            # if a trackerhandler is set on this object, update it with the
        #            # fieldname and the new value
        #            fieldtracker = "_oob_at_%s_postsave" % fieldname
        #            if hasattr(self, fieldtracker):
        #                _GA(self, fieldtracker)(fieldname)
        pass


class WeakSharedMemoryModelBase(SharedMemoryModelBase):
    """
    Uses a WeakValue dictionary for caching instead of a regular one.

    """

    def _prepare(cls):
        super()._prepare()
        cls.__dbclass__.__instance_cache__ = WeakValueDictionary()


class WeakSharedMemoryModel(SharedMemoryModel, metaclass=WeakSharedMemoryModelBase):
    """
    Uses a WeakValue dictionary for caching instead of a regular one

    """

    class Meta(object):
        abstract = True


def flush_cache(**kwargs):
    """
    Flush idmapper cache. When doing so the cache will fire the
    at_idmapper_flush hook to allow the object to optionally handle
    its own flushing.

    Uses a signal so we make sure to catch cascades.

    """

    def class_hierarchy(clslist):
        """Recursively yield a class hierarchy"""
        for cls in clslist:
            subclass_list = cls.__subclasses__()
            if subclass_list:
                for subcls in class_hierarchy(subclass_list):
                    yield subcls
            else:
                yield cls

    for cls in class_hierarchy([SharedMemoryModel]):
        cls.flush_instance_cache()
    # run the python garbage collector
    return gc.collect()


# request_finished.connect(flush_cache)
post_migrate.connect(flush_cache)


def flush_cached_instance(sender, instance, **kwargs):
    """
    Flush the idmapper cache only for a given instance.

    """
    # XXX: Is this the best way to make sure we can flush?
    if not hasattr(instance, "flush_cached_instance"):
        return
    sender.flush_cached_instance(instance, force=True)


pre_delete.connect(flush_cached_instance)


def update_cached_instance(sender, instance, **kwargs):
    """
    Re-cache the given instance in the idmapper cache.

    """
    if not hasattr(instance, "cache_instance"):
        return
    sender.cache_instance(instance)


post_save.connect(update_cached_instance)


LAST_FLUSH = None


def conditional_flush(max_rmem, force=False):
    """
    Flush the cache if the estimated memory usage exceeds `max_rmem`.

    The flusher has a timeout to avoid flushing over and over
    in particular situations (this means that for some setups
    the memory usage will exceed the requirement and a server with
    more memory is probably required for the given game).

    Args:
        max_rmem (int): memory-usage estimation-treshold after which
            cache is flushed.
        force (bool, optional): forces a flush, regardless of timeout.
            Defaults to `False`.

    """
    global LAST_FLUSH

    def mem2cachesize(desired_rmem):
        """
        Estimate the size of the idmapper cache based on the memory
        desired. This is used to optionally cap the cache size.

        desired_rmem - memory in MB (minimum 50MB)

        The formula is empirically estimated from usage tests (Linux)
        and is
            Ncache = RMEM - 35.0 / 0.0157
        where RMEM is given in MB and Ncache is the size of the cache
        for this memory usage. VMEM tends to be about 100MB higher
        than RMEM for large memory usage.
        """
        vmem = max(desired_rmem, 50.0)
        Ncache = int(abs(float(vmem) - 35.0) / 0.0157)
        return Ncache

    if not max_rmem:
        # auto-flush is disabled
        return

    now = time.time()
    if not LAST_FLUSH:
        # server is just starting
        LAST_FLUSH = now
        return

    if ((now - LAST_FLUSH) < AUTO_FLUSH_MIN_INTERVAL) and not force:
        # too soon after last flush.
        logger.log_warn(
            "Warning: Idmapper flush called more than once in %s min interval. Check memory usage."
            % (AUTO_FLUSH_MIN_INTERVAL / 60.0)
        )
        return

    if os.name == "nt":
        # we can't look for mem info in Windows at the moment
        return

    # check actual memory usage
    Ncache_max = mem2cachesize(max_rmem)
    Ncache, _ = cache_size()
    actual_rmem = (
        float(os.popen("ps -p %d -o %s | tail -1" % (os.getpid(), "rss")).read()) / 1000.0
    )  # resident memory

    if Ncache >= Ncache_max and actual_rmem > max_rmem * 0.9:
        # flush cache when number of objects in cache is big enough and our
        # actual memory use is within 10% of our set max
        flush_cache()
        LAST_FLUSH = now


def cache_size(mb=True):
    """
    Calculate statistics about the cache.

    Note: we cannot get reliable memory statistics from the cache -
    whereas we could do `getsizof` each object in cache, the result is
    highly imprecise and for a large number of objects the result is
    many times larger than the actual memory usage of the entire server;
    Python is clearly reusing memory behind the scenes that we cannot
    catch in an easy way here.  Ideas are appreciated. /Griatch

    Returns:
      total_num, {objclass:total_num, ...}

    """
    numtotal = [0]  # use mutable to keep reference through recursion
    classdict = {}

    def get_recurse(submodels):
        for submodel in submodels:
            subclasses = submodel.__subclasses__()
            if not subclasses:
                num = len(submodel.get_all_cached_instances())
                numtotal[0] += num
                classdict[submodel.__dbclass__.__name__] = num
            else:
                get_recurse(subclasses)

    get_recurse(SharedMemoryModel.__subclasses__())
    return numtotal[0], classdict



"""
This sub-package holds the miscelaneous utilities used by other
modules in Evennia. It also holds the idmapper in-memory caching
functionality.

"""

# simple check to determine if we are currently running under pypy.
try:
    import __pypy__ as is_pypy
except ImportError:
    is_pypy = False

from .utils import *



"""
This implements the common managers that are used by the
abstract models in dbobjects.py (and which are thus shared by
all Attributes and TypedObjects).

"""
import shlex

from django.db.models import Count, ExpressionWrapper, F, FloatField, Q
from django.db.models.functions import Cast

from evennia.typeclasses.attributes import Attribute
from evennia.typeclasses.tags import Tag
from evennia.utils import idmapper
from evennia.utils.utils import class_from_module, make_iter, variable_from_module

__all__ = ("TypedObjectManager",)
_GA = object.__getattribute__
_Tag = None


# Managers


class TypedObjectManager(idmapper.manager.SharedMemoryManager):
    """
    Common ObjectManager for all dbobjects.

    """

    # common methods for all typed managers. These are used
    # in other methods. Returns querysets.

    # Attribute manager methods
    def get_attribute(
        self, key=None, category=None, value=None, strvalue=None, obj=None, attrtype=None, **kwargs
    ):
        """
        Return Attribute objects by key, by category, by value, by strvalue, by
        object (it is stored on) or with a combination of those criteria.

        Args:
            key (str, optional): The attribute's key to search for
            category (str, optional): The category of the attribute(s) to search for.
            value (str, optional): The attribute value to search for.
                Note that this is not a very efficient operation since it
                will query for a pickled entity. Mutually exclusive to
                `strvalue`.
            strvalue (str, optional): The str-value to search for.
                Most Attributes will not have strvalue set. This is
                mutually exclusive to the `value` keyword and will take
                precedence if given.
            obj (Object, optional): On which object the Attribute to
                search for is.
            attrype (str, optional): An attribute-type to search for.
                By default this is either `None` (normal Attributes) or
                `"nick"`.
            **kwargs (any): Currently unused. Reserved for future use.

        Returns:
            list: The matching Attributes.

        """
        dbmodel = self.model.__dbclass__.__name__.lower()
        query = [("attribute__db_attrtype", attrtype), ("attribute__db_model", dbmodel)]
        if obj:
            query.append(("%s__id" % self.model.__dbclass__.__name__.lower(), obj.id))
        if key:
            query.append(("attribute__db_key", key))
        if category:
            query.append(("attribute__db_category", category))
        if strvalue:
            query.append(("attribute__db_strvalue", strvalue))
        if value:
            # no reason to make strvalue/value mutually exclusive at this level
            query.append(("attribute__db_value", value))
        return Attribute.objects.filter(
            pk__in=self.model.db_attributes.through.objects.filter(**dict(query)).values_list(
                "attribute_id", flat=True
            )
        )

    def get_nick(self, key=None, category=None, value=None, strvalue=None, obj=None):
        """
        Get a nick, in parallel to `get_attribute`.

        Args:
            key (str, optional): The nicks's key to search for
            category (str, optional): The category of the nicks(s) to search for.
            value (str, optional): The attribute value to search for. Note that this
                is not a very efficient operation since it will query for a pickled
                entity. Mutually exclusive to `strvalue`.
            strvalue (str, optional): The str-value to search for. Most Attributes
                will not have strvalue set. This is mutually exclusive to the `value`
                keyword and will take precedence if given.
            obj (Object, optional): On which object the Attribute to search for is.

        Returns:
            nicks (list): The matching Nicks.

        """
        return self.get_attribute(
            key=key, category=category, value=value, strvalue=strvalue, obj=obj
        )

    def get_by_attribute(
        self, key=None, category=None, value=None, strvalue=None, attrtype=None, **kwargs
    ):
        """
        Return objects having attributes with the given key, category,
        value, strvalue or combination of those criteria.

        Args:
            key (str, optional): The attribute's key to search for
            category (str, optional): The category of the attribute
                to search for.
            value (str, optional): The attribute value to search for.
                Note that this is not a very efficient operation since it
                will query for a pickled entity. Mutually exclusive to
                `strvalue`.
            strvalue (str, optional): The str-value to search for.
                Most Attributes will not have strvalue set. This is
                mutually exclusive to the `value` keyword and will take
                precedence if given.
            attrype (str, optional): An attribute-type to search for.
                By default this is either `None` (normal Attributes) or
                `"nick"`.
            kwargs (any): Currently unused. Reserved for future use.

        Returns:
            obj (list): Objects having the matching Attributes.

        """
        dbmodel = self.model.__dbclass__.__name__.lower()
        query = [
            ("db_attributes__db_attrtype", attrtype),
            ("db_attributes__db_model", dbmodel),
        ]
        if key:
            query.append(("db_attributes__db_key", key))
        if category:
            query.append(("db_attributes__db_category", category))
        if strvalue:
            query.append(("db_attributes__db_strvalue", strvalue))
        elif value:
            # strvalue and value are mutually exclusive
            query.append(("db_attributes__db_value", value))
        return self.filter(**dict(query))

    def get_by_nick(self, key=None, nick=None, category="inputline"):
        """
        Get object based on its key or nick.

        Args:
            key (str, optional): The attribute's key to search for
            nick (str, optional): The nickname to search for
            category (str, optional): The category of the nick
                to search for.

        Returns:
            obj (list): Objects having the matching Nicks.

        """
        return self.get_by_attribute(key=key, category=category, strvalue=nick, attrtype="nick")

    # Tag manager methods

    def get_tag(self, key=None, category=None, obj=None, tagtype=None, global_search=False):
        """
        Return Tag objects by key, by category, by object (it is
        stored on) or with a combination of those criteria.

        Args:
            key (str, optional): The Tag's key to search for
            category (str, optional): The Tag of the attribute(s)
                to search for.
            obj (Object, optional): On which object the Tag to
                search for is.
            tagtype (str, optional): One of `None` (normal tags),
                "alias" or "permission"
            global_search (bool, optional): Include all possible tags,
                not just tags on this object

        Returns:
            tag (list): The matching Tags.

        """
        global _Tag
        if not _Tag:
            from evennia.typeclasses.models import Tag as _Tag
        dbmodel = self.model.__dbclass__.__name__.lower()
        if global_search:
            # search all tags using the Tag model
            query = [("db_tagtype", tagtype), ("db_model", dbmodel)]
            if obj:
                query.append(("id", obj.id))
            if key:
                query.append(("db_key", key))
            if category:
                query.append(("db_category", category))
            else:
                query.append(("db_category", None))
            return _Tag.objects.filter(**dict(query))
        else:
            # search only among tags stored on on this model
            query = [("tag__db_tagtype", tagtype), ("tag__db_model", dbmodel)]
            if obj:
                query.append(("%s__id" % self.model.__name__.lower(), obj.id))
            if key:
                query.append(("tag__db_key", key))
            if category:
                query.append(("tag__db_category", category))
            return Tag.objects.filter(
                pk__in=self.model.db_tags.through.objects.filter(**dict(query)).values_list(
                    "tag_id", flat=True
                )
            )

    def get_permission(self, key=None, category=None, obj=None):
        """
        Get a permission from the database.

        Args:
            key (str, optional): The permission's identifier.
            category (str, optional): The permission's category.
            obj (object, optional): The object on which this Tag is set.

        Returns:
            permission (list): Permission objects.

        """
        return self.get_tag(key=key, category=category, obj=obj, tagtype="permission")

    def get_alias(self, key=None, category=None, obj=None):
        """
        Get an alias from the database.

        Args:
            key (str, optional): The permission's identifier.
            category (str, optional): The permission's category.
            obj (object, optional): The object on which this Tag is set.

        Returns:
            alias (list): Alias objects.

        """
        return self.get_tag(key=key, category=category, obj=obj, tagtype="alias")

    def get_by_tag(self, key=None, category=None, tagtype=None, **kwargs):
        """
        Return objects having tags with a given key or category or combination of the two.
        Also accepts multiple tags/category/tagtype

        Args:
            key (str or list, optional): Tag key or list of keys. Not case sensitive.
            category (str or list, optional): Tag category. Not case sensitive.
                If `key` is a list, a single category can either apply to all
                keys in that list or this must be a list matching the `key`
                list element by element. If no `key` is given, all objects with
                tags of this category are returned.
            tagtype (str, optional): 'type' of Tag, by default
                this is either `None` (a normal Tag), `alias` or
                `permission`. This always apply to all queried tags.

        Keyword Args:
            match (str): "all" (default) or "any"; determines whether the
                target object must be tagged with ALL of the provided
                tags/categories or ANY single one. ANY will perform a weighted
                sort, so objects with more tag matches will outrank those with
                fewer tag matches.

        Returns:
            objects (list): Objects with matching tag.

        Raises:
            IndexError: If `key` and `category` are both lists and `category` is shorter
                than `key`.

        """
        if not (key or category):
            return []

        global _Tag
        if not _Tag:
            from evennia.typeclasses.models import Tag as _Tag

        anymatch = "any" == kwargs.get("match", "all").lower().strip()

        keys = make_iter(key) if key else []
        categories = make_iter(category) if category else []
        n_keys = len(keys)
        n_categories = len(categories)
        unique_categories = set(categories)
        n_unique_categories = len(unique_categories)

        dbmodel = self.model.__dbclass__.__name__.lower()
        query = (
            self.filter(db_tags__db_tagtype__iexact=tagtype, db_tags__db_model__iexact=dbmodel)
            .distinct()
            .order_by("id")
        )

        if n_keys > 0:
            # keys and/or categories given
            if n_categories == 0:
                categories = [None for _ in range(n_keys)]
            elif n_categories == 1 and n_keys > 1:
                cat = categories[0]
                categories = [cat for _ in range(n_keys)]
            elif 1 < n_categories < n_keys:
                raise IndexError(
                    "get_by_tag needs a single category or a list of categories "
                    "the same length as the list of tags."
                )
            clauses = Q()
            for ikey, key in enumerate(keys):
                # ANY mode; must match any one of the given tags/categories
                clauses |= Q(db_key__iexact=key, db_category__iexact=categories[ikey])
        else:
            # only one or more categories given
            clauses = Q()
            # ANY mode; must match any one of them
            for category in unique_categories:
                clauses |= Q(db_category__iexact=category)

        tags = _Tag.objects.filter(clauses)
        query = query.filter(db_tags__in=tags).annotate(
            matches=Count("db_tags__pk", filter=Q(db_tags__in=tags), distinct=True)
        )

        if anymatch:
            # ANY: Match any single tag, ordered by weight
            query = query.order_by("-matches")
        else:
            # Default ALL: Match all of the tags and optionally more
            n_req_tags = n_keys if n_keys > 0 else n_unique_categories
            query = query.filter(matches__gte=n_req_tags)

        return query

    def get_by_permission(self, key=None, category=None):
        """
        Return objects having permissions with a given key or category or
        combination of the two.

        Args:
            key (str, optional): Permissions key. Not case sensitive.
            category (str, optional): Permission category. Not case sensitive.
        Returns:
            objects (list): Objects with matching permission.
        """
        return self.get_by_tag(key=key, category=category, tagtype="permission")

    def get_by_alias(self, key=None, category=None):
        """
        Return objects having aliases with a given key or category or
        combination of the two.

        Args:
            key (str, optional): Alias key. Not case sensitive.
            category (str, optional): Alias category. Not case sensitive.
        Returns:
            objects (list): Objects with matching alias.
        """
        return self.get_by_tag(key=key, category=category, tagtype="alias")

    def create_tag(self, key=None, category=None, data=None, tagtype=None):
        """
        Create a new Tag of the base type associated with this
        object.  This makes sure to create case-insensitive tags.
        If the exact same tag configuration (key+category+tagtype+dbmodel)
        exists on the model, a new tag will not be created, but an old
        one returned.


        Args:
            key (str, optional): Tag key. Not case sensitive.
            category (str, optional): Tag category. Not case sensitive.
            data (str, optional): Extra information about the tag.
            tagtype (str or None, optional): 'type' of Tag, by default
                this is either `None` (a normal Tag), `alias` or
                `permission`.
        Notes:
            The `data` field is not part of the uniqueness of the tag:
            Setting `data` on an existing tag will overwrite the old
            data field. It is intended only as a way to carry
            information about the tag (like a help text), not to carry
            any information about the tagged objects themselves.

        """
        data = str(data) if data is not None else None
        # try to get old tag

        dbmodel = self.model.__dbclass__.__name__.lower()
        tag = self.get_tag(key=key, category=category, tagtype=tagtype, global_search=True)
        if tag and data is not None:
            # get tag from list returned by get_tag
            tag = tag[0]
            # overload data on tag
            tag.db_data = data
            tag.save()
        elif not tag:
            # create a new tag
            global _Tag
            if not _Tag:
                from evennia.typeclasses.models import Tag as _Tag
            tag = _Tag.objects.create(
                db_key=key.strip().lower() if key is not None else None,
                db_category=category.strip().lower() if category and key is not None else None,
                db_data=data,
                db_model=dbmodel,
                db_tagtype=tagtype.strip().lower() if tagtype is not None else None,
            )
            tag.save()
        return make_iter(tag)[0]

    def dbref(self, dbref, reqhash=True):
        """
        Determing if input is a valid dbref.

        Args:
            dbref (str or int): A possible dbref.
            reqhash (bool, optional): If the "#" is required for this
                to be considered a valid hash.

        Returns:
            dbref (int or None): The integer part of the dbref.

        Notes:
            Valid forms of dbref (database reference number) are
            either a string '#N' or an integer N.

        """
        if reqhash and not (isinstance(dbref, str) and dbref.startswith("#")):
            return None
        if isinstance(dbref, str):
            dbref = dbref.lstrip("#")
        try:
            if int(dbref) < 0:
                return None
        except Exception:
            return None
        return dbref

    def get_id(self, dbref):
        """
        Find object with given dbref.

        Args:
            dbref (str or int): The id to search for.

        Returns:
            object (TypedObject): The matched object.

        """
        dbref = self.dbref(dbref, reqhash=False)
        try:
            return self.get(id=dbref)
        except self.model.DoesNotExist:
            pass
        return None

    def dbref_search(self, dbref):
        """
        Alias to get_id.

        Args:
            dbref (str or int): The id to search for.

        Returns:
            Queryset: Queryset with 0 or 1 match.

        """
        dbref = self.dbref(dbref, reqhash=False)
        if dbref:
            return self.filter(id=dbref)
        return self.none()

    search_dbref = dbref_search  # alias

    def get_dbref_range(self, min_dbref=None, max_dbref=None):
        """
        Get objects within a certain range of dbrefs.

        Args:
            min_dbref (int): Start of dbref range.
            max_dbref (int): End of dbref range (inclusive)

        Returns:
            objects (list): TypedObjects with dbrefs within
                the given dbref ranges.

        """
        retval = super().all()
        if min_dbref is not None:
            retval = retval.filter(id__gte=self.dbref(min_dbref, reqhash=False))
        if max_dbref is not None:
            retval = retval.filter(id__lte=self.dbref(max_dbref, reqhash=False))
        return retval

    def get_typeclass_totals(self, *args, **kwargs) -> object:
        """
        Returns a queryset of typeclass composition statistics.

        Returns:
            qs (Queryset): A queryset of dicts containing the typeclass (name),
                the count of objects with that typeclass and a float representing
                the percentage of objects associated with the typeclass.

        """
        return (
            self.values("db_typeclass_path")
            .distinct()
            .annotate(
                # Get count of how many objects for each typeclass exist
                count=Count("db_typeclass_path")
            )
            .annotate(
                # Rename db_typeclass_path field to something more human
                typeclass=F("db_typeclass_path"),
                # Calculate this class' percentage of total composition
                percent=ExpressionWrapper(
                    ((F("count") / float(self.count())) * 100.0),
                    output_field=FloatField(),
                ),
            )
            .values("typeclass", "count", "percent")
        )

    def object_totals(self):
        """
        Get info about database statistics.

        Returns:
            census (dict): A dictionary `{typeclass_path: number, ...}` with
                all the typeclasses active in-game as well as the number
                of such objects defined (i.e. the number of database
                object having that typeclass set on themselves).

        """
        stats = self.get_typeclass_totals().order_by("typeclass")
        return {x.get("typeclass"): x.get("count") for x in stats}

    def typeclass_search(self, typeclass, include_children=False, include_parents=False):
        """
        Searches through all objects returning those which are of the
        specified typeclass.

        Args:
            typeclass (str or class): A typeclass class or a python path to a typeclass.
            include_children (bool, optional): Return objects with
                given typeclass *and* all children inheriting from this
                typeclass. Mutually exclusive to `include_parents`.
            include_parents (bool, optional): Return objects with
                given typeclass *and* all parents to this typeclass.
                Mutually exclusive to `include_children`.

        Returns:
            objects (list): The objects found with the given typeclasses.

        Raises:
            ImportError: If the provided `typeclass` is not a valid typeclass or the
                path to an existing typeclass.

        """
        if not callable(typeclass):
            typeclass = class_from_module(typeclass)

        if include_children:
            query = typeclass.objects.all_family()
        else:
            query = typeclass.objects.all()

        if include_parents:
            parents = typeclass.__mro__
            if parents:
                parent_queries = []
                for parent in (parent for parent in parents if hasattr(parent, "path")):
                    parent_queries.append(super().filter(db_typeclass_path__exact=parent.path))
                query = query.union(*parent_queries)

        return query


class TypeclassManager(TypedObjectManager):
    """
    Manager for the typeclasses. The main purpose of this manager is
    to limit database queries to the given typeclass despite all
    typeclasses technically being defined in the same core database
    model.

    """

    # object-manager methods
    def smart_search(self, query):
        """
        Search by supplying a string with optional extra search criteria to aid the query.

        Args:
            query (str): A search criteria that accepts extra search criteria on the following
            forms:

                [key|alias|#dbref...]
                [tag==<tagstr>[:category]...]
                [attr==<key>:<value>:category...]

            All three can be combined in the same query, separated by spaces.

        Returns:
            matches (queryset): A queryset result matching all queries exactly. If wanting to use
                spaces or ==, != in tags or attributes, enclose them in quotes.

        Example:
            house = smart_search("key=foo alias=bar tag=house:building tag=magic attr=color:red")

        Note:
            The flexibility of this method is limited by the input line format. Tag/attribute
            matching only works for matching primitives.  For even more complex queries, such as
            'in' operations or object field matching, use the full django query language.

        """
        # shlex splits by spaces unless escaped by quotes
        querysplit = shlex.split(query)
        queries, plustags, plusattrs, negtags, negattrs = [], [], [], [], []
        for ipart, part in enumerate(querysplit):
            key, rest = part, ""
            if ":" in part:
                key, rest = part.split(":", 1)
            # tags are on the form tag or tag:category
            if key.startswith("tag=="):
                plustags.append((key[5:], rest))
                continue
            elif key.startswith("tag!="):
                negtags.append((key[5:], rest))
                continue
            # attrs are on the form attr:value or attr:value:category
            elif rest:
                value, category = rest, ""
                if ":" in rest:
                    value, category = rest.split(":", 1)
                if key.startswith("attr=="):
                    plusattrs.append((key[7:], value, category))
                    continue
                elif key.startswith("attr!="):
                    negattrs.append((key[7:], value, category))
                    continue
            # if we get here, we are entering a key search criterion which
            # we assume is one word.
            queries.append(part)
        # build query from components
        query = " ".join(queries)
        # TODO

    def get(self, *args, **kwargs):
        """
        Overload the standard get. This will limit itself to only
        return the current typeclass.

        Args:
            args (any): These are passed on as arguments to the default
                django get method.
        Keyword Args:
            kwargs (any): These are passed on as normal arguments
                to the default django get method
        Returns:
            object (object): The object found.

        Raises:
            ObjectNotFound: The exact name of this exception depends
                on the model base used.

        """
        kwargs.update({"db_typeclass_path": self.model.path})
        return super().get(**kwargs)

    def filter(self, *args, **kwargs):
        """
        Overload of the standard filter function. This filter will
        limit itself to only the current typeclass.

        Args:
            args (any): These are passed on as arguments to the default
                django filter method.
        Keyword Args:
            kwargs (any): These are passed on as normal arguments
                to the default django filter method.
        Returns:
            objects (queryset): The objects found.

        """
        kwargs.update({"db_typeclass_path": self.model.path})
        return super().filter(*args, **kwargs)

    def all(self):
        """
        Overload method to return all matches, filtering for typeclass.

        Returns:
            objects (queryset): The objects found.

        """
        return super().all().filter(db_typeclass_path=self.model.path)

    def first(self):
        """
        Overload method to return first match, filtering for typeclass.

        Returns:
            object (object): The object found.

        Raises:
            ObjectNotFound: The exact name of this exception depends
                on the model base used.

        """
        return super().filter(db_typeclass_path=self.model.path).first()

    def last(self):
        """
        Overload method to return last match, filtering for typeclass.

        Returns:
            object (object): The object found.

        Raises:
            ObjectNotFound: The exact name of this exception depends
                on the model base used.

        """
        return super().filter(db_typeclass_path=self.model.path).last()

    def count(self):
        """
        Overload method to return number of matches, filtering for typeclass.

        Returns:
            integer : Number of objects found.

        """
        return super().filter(db_typeclass_path=self.model.path).count()

    def annotate(self, *args, **kwargs):
        """
        Overload annotate method to filter on typeclass before annotating.
        Args:
            *args (any): Positional arguments passed along to queryset annotate method.
            **kwargs (any): Keyword arguments passed along to queryset annotate method.

        Returns:
            Annotated queryset.
        """
        return super().filter(db_typeclass_path=self.model.path).annotate(*args, **kwargs)

    def values(self, *args, **kwargs):
        """
        Overload values method to filter on typeclass first.
        Args:
            *args (any): Positional arguments passed along to values method.
            **kwargs (any): Keyword arguments passed along to values method.

        Returns:
            Queryset of values dictionaries, just filtered by typeclass first.
        """
        return super().filter(db_typeclass_path=self.model.path).values(*args, **kwargs)

    def values_list(self, *args, **kwargs):
        """
        Overload values method to filter on typeclass first.
        Args:
            *args (any): Positional arguments passed along to values_list method.
            **kwargs (any): Keyword arguments passed along to values_list method.

        Returns:
            Queryset of value_list tuples, just filtered by typeclass first.
        """
        return super().filter(db_typeclass_path=self.model.path).values_list(*args, **kwargs)

    def _get_subclasses(self, cls):
        """
        Recursively get all subclasses to a class.

        Args:
            cls (classoject): A class to get subclasses from.
        """
        all_subclasses = cls.__subclasses__()
        for subclass in all_subclasses:
            all_subclasses.extend(self._get_subclasses(subclass))
        return all_subclasses

    def get_family(self, *args, **kwargs):
        """
        Variation of get that not only returns the current typeclass
        but also all subclasses of that typeclass.

        Keyword Args:
            kwargs (any): These are passed on as normal arguments
                to the default django get method.
        Returns:
            objects (list): The objects found.

        Raises:
            ObjectNotFound: The exact name of this exception depends
                on the model base used.

        """
        paths = [self.model.path] + [
            "%s.%s" % (cls.__module__, cls.__name__) for cls in self._get_subclasses(self.model)
        ]
        kwargs.update({"db_typeclass_path__in": paths})
        return super().get(*args, **kwargs)

    def filter_family(self, *args, **kwargs):
        """
        Variation of filter that allows results both from typeclass
        and from subclasses of typeclass

        Args:
            args (any): These are passed on as arguments to the default
                django filter method.
        Keyword Args:
            kwargs (any): These are passed on as normal arguments
                to the default django filter method.
        Returns:
            objects (list): The objects found.

        """
        # query, including all subclasses
        paths = [self.model.path] + [
            "%s.%s" % (cls.__module__, cls.__name__) for cls in self._get_subclasses(self.model)
        ]
        kwargs.update({"db_typeclass_path__in": paths})
        return super().filter(*args, **kwargs)

    def all_family(self):
        """
        Return all matches, allowing matches from all subclasses of
        the typeclass.

        Returns:
            objects (list): The objects found.

        """
        paths = [self.model.path] + [
            "%s.%s" % (cls.__module__, cls.__name__) for cls in self._get_subclasses(self.model)
        ]
        return super().all().filter(db_typeclass_path__in=paths)



"""
Prototypes

A prototype is a simple way to create individualized instances of a
given typeclass. It is dictionary with specific key names.

For example, you might have a Sword typeclass that implements everything a
Sword would need to do. The only difference between different individual Swords
would be their key, description and some Attributes. The Prototype system
allows to create a range of such Swords with only minor variations. Prototypes
can also inherit and combine together to form entire hierarchies (such as
giving all Sabres and all Broadswords some common properties). Note that bigger
variations, such as custom commands or functionality belong in a hierarchy of
typeclasses instead.

A prototype can either be a dictionary placed into a global variable in a
python module (a 'module-prototype') or stored in the database as a dict on a
special Script (a db-prototype). The former can be created just by adding dicts
to modules Evennia looks at for prototypes, the latter is easiest created
in-game via the `olc` command/menu.

Prototypes are read and used to create new objects with the `spawn` command
or directly via `evennia.spawn` or the full path `evennia.prototypes.spawner.spawn`.

A prototype dictionary have the following keywords:

Possible keywords are:
- `prototype_key` - the name of the prototype. This is required for db-prototypes,
  for module-prototypes, the global variable name of the dict is used instead
- `prototype_parent` - string pointing to parent prototype if any. Prototype inherits
  in a similar way as classes, with children overriding values in their parents.
- `key` - string, the main object identifier.
- `typeclass` - string, if not set, will use `settings.BASE_OBJECT_TYPECLASS`.
- `location` - this should be a valid object or #dbref.
- `home` - valid object or #dbref.
- `destination` - only valid for exits (object or #dbref).
- `permissions` - string or list of permission strings.
- `locks` - a lock-string to use for the spawned object.
- `aliases` - string or list of strings.
- `attrs` - Attributes, expressed as a list of tuples on the form `(attrname, value)`,
  `(attrname, value, category)`, or `(attrname, value, category, locks)`. If using one
   of the shorter forms, defaults are used for the rest.
- `tags` - Tags, as a list of tuples `(tag,)`, `(tag, category)` or `(tag, category, data)`.
-  Any other keywords are interpreted as Attributes with no category or lock.
   These will internally be added to `attrs` (equivalent to `(attrname, value)`.

See the `spawn` command and `evennia.prototypes.spawner.spawn` for more info.

"""

## example of module-based prototypes using
## the variable name as `prototype_key` and
## simple Attributes

# from random import randint
#
# GOBLIN = {
# "key": "goblin grunt",
# "health": lambda: randint(20,30),
# "resists": ["cold", "poison"],
# "attacks": ["fists"],
# "weaknesses": ["fire", "light"],
# "tags": = [("greenskin", "monster"), ("humanoid", "monster")]
# }
#
# GOBLIN_WIZARD = {
# "prototype_parent": "GOBLIN",
# "key": "goblin wizard",
# "spells": ["fire ball", "lighting bolt"]
# }
#
# GOBLIN_ARCHER = {
# "prototype_parent": "GOBLIN",
# "key": "goblin archer",
# "attacks": ["short bow"]
# }
#
# This is an example of a prototype without a prototype
# (nor key) of its own, so it should normally only be
# used as a mix-in, as in the example of the goblin
# archwizard below.
# ARCHWIZARD_MIXIN = {
# "attacks": ["archwizard staff"],
# "spells": ["greater fire ball", "greater lighting"]
# }
#
# GOBLIN_ARCHWIZARD = {
# "key": "goblin archwizard",
# "prototype_parent" : ("GOBLIN_WIZARD", "ARCHWIZARD_MIXIN")
# }
